# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2022.                 #
# ######################################### #

from contextlib import contextmanager
from functools import wraps
from typing import Optional, Tuple, Type, Union


class _RetryableError(Exception):
    def __init__(self, original_error: Exception):
        self.message = ''
        self.original_error = original_error
        self.retries = None

    def __str__(self):
        return self.message


def retry(
        on: Optional[Union[Type[Exception], Tuple[Type[Exception]]]] = None,
        n_times: int = 3,
):
    """
    A decorator to be used on a flaky test. The test will be rerun until no
    exception defined in `on` is thrown, up to `n_times`. If `on` is
    unspecified, then the functionality of `flaky_assertions` is expected to be
    applied: the test will only be retried on AssertionErrors occurring within
    a `with flaky_assertions(): ...` statement.
    """
    if not on:
        on = _RetryableError

    def decorator(test_function):
        @wraps(test_function)
        def wrapper(*args, **kwargs):
            for i in range(n_times):
                try:
                    return test_function(*args, **kwargs)
                except on as e:
                    if i + 1 < n_times:
                        continue

                    if isinstance(e, _RetryableError):
                        # This is maybe not the prettiest, but let's hide
                        # the traceback as it pollutes the pytest output,
                        # and detracts from the traceback of the actual
                        # error:
                        e.__traceback__ = None
                        e.message = (f'Failing on retry number {i + 1}. '
                                     f'Raising the original error now!')

                        raise e.original_error
                    raise e
        return wrapper
    return decorator


@contextmanager
def flaky_assertions():
    try:
        yield
    except AssertionError as e:
        raise _RetryableError(original_error=e)



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

def generate_2D_gaussian(num_particles):

    '''
    Generate a 2D Gaussian distribution.

    Parameters
    ----------
    num_particles : int
        Number of particles to be generated.

    Returns
    -------
    x1 : np.ndarray
        First normalized coordinate.
    x2 : np.ndarray
        Second normalized coordinate.

    '''

    x_norm = np.random.normal(size=num_particles)
    px_norm = np.random.normal(size=num_particles)

    return x_norm, px_norm




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt

from xobjects.test_helpers import for_all_test_contexts

test_data_folder = xt._pkg_root.joinpath('../test_data').absolute()


@for_all_test_contexts
def test_phase_monitor(test_context):
    filename = test_data_folder.joinpath(
        'sps_w_spacecharge/line_no_spacecharge_and_particle.json')
    with open(filename, 'r') as fid:
        ddd = json.load(fid)
    line = xt.Line.from_dict(ddd['line'])
    line.particle_ref = xp.Particles.from_dict(ddd['particle'])

    line.build_tracker(_context=test_context)

    particles = xp.build_particles(line=line, x_norm=[0.1, 0.2],
                                   y_norm=[0.3, 0.4],
                                   nemitt_x=2e-6, nemitt_y=2e-6,
                                   _context=test_context)
    phase_monitor = xp.PhaseMonitor(line=line, num_particles=2,
                                    twiss=line.twiss())

    for _ in range(5):
        phase_monitor.measure(particles)
        line.track(particles)

    tw = line.twiss()
    xo.assert_allclose(phase_monitor.qx[:], np.mod(tw['qx'], 1),
                       rtol=0, atol=1e-3)
    xo.assert_allclose(phase_monitor.qy[:], np.mod(tw['qy'], 1),
                       rtol=0, atol=1e-3)





import numpy as np




def generate_hypersphere(N, D, r=1, rng_seed = 0, surface=False ,unpack = False):
    '''
    Generate points uniformly distributed inside or on the surface of an N-dimensional hypersphere.
    Adapted from : https://baezortega.github.io/2018/10/14/hypersphere-sampling/

    Parameters
    ----------
    N : int
        Number of particles to generate.
    D : int
        Dimension of the hypersphere.
    r : float or list
        Radius of the hypersphere. If a list, specifies radii for anisotropic scaling.
    rng_seed : int
        Seed for the random number generator for reproducibility.
    surface : bool
        If True, points will be generated on the surface. If False, points will be generated inside the hypersphere.
    unpack : bool
        If True, returns individual arrays for each dimension. If False, returns a single array with shape (N, D).

    Returns
    -------
    samples : np.ndarray or tuple of np.ndarray
        Generated points. Shape is (N, D) if unpack is False, otherwise D arrays of shape (N,).
    '''
    # Set the random seed for reproducibility
    rng = np.random.default_rng(int(rng_seed))

    # Sample D vectors of N Gaussian coordinates
    N = int(N)
    D = int(D)
    samples = rng.standard_normal(size = (N, D))

    # Normalise all distances (radii) to 1
    radii = np.sqrt(np.sum(samples ** 2, axis=1))[:,np.newaxis]
    samples = samples / radii

    # Sample N radii with exponential distribution (unless points are to be on the surface)
    if not surface:
        new_radii = np.random.uniform(low=0.0, high=1.0, size=(N, 1)) ** (1 / D)
        samples = samples * new_radii

    # Scale the samples to the desired radius
    if isinstance(r,list):
        r = np.array(r)[np.newaxis,:]
    elif isinstance(r,type(np.array([]))):
        assert False, 'r should be float or list'
    samples = samples * r

    if not unpack:
        return samples
    else:
        return samples.T


def generate_hypersphere_2D(num_particles,r = 1, rng_seed = 0):
    '''
    Generate points uniformly distributed inside a 2-dimensional hypersphere (circle).

    Parameters
    ----------
    num_particles : int
        Number of particles to generate.
    r : float
        Radius of the circle.
    rng_seed : int
        Seed for the random number generator for reproducibility.

    Returns
    -------
    x_norm : np.ndarray
        x-coordinates of the generated points.
    px_norm : np.ndarray
        y-coordinates of the generated points.
    '''
    x_norm , px_norm  = generate_hypersphere(num_particles,D=2,r=r, rng_seed=rng_seed, surface = False,unpack=True)

    return x_norm, px_norm


def generate_hypersphere_4D(num_particles,rx =1,ry =1, rng_seed = 0):
    '''
    Generate points uniformly distributed inside a 4-dimensional hypersphere with anisotropic scaling.

    Parameters
    ----------
    num_particles : int
        Number of particles to generate.
    rx : float
        Scaling factor for the x and px dimensions.
    ry : float
        Scaling factor for the y and py dimensions.
    rng_seed : int
        Seed for the random number generator for reproducibility.

    Returns
    -------
    x_norm, px_norm, y_norm, py_norm : np.ndarray
        Coordinates of the generated points in the 4-dimensional space.
    '''

    x_norm , px_norm , y_norm, py_norm = generate_hypersphere(num_particles,D=4,r=[rx,rx,ry,ry], rng_seed=rng_seed, surface = False,unpack=True)

    return x_norm , px_norm , y_norm, py_norm


def generate_hypersphere_6D(num_particles,rx =1,ry =1, rzeta=1, rng_seed = 0):
    '''
    Generate points uniformly distributed inside a 6-dimensional hypersphere with anisotropic scaling.

    Parameters
    ----------
    num_particles : int
        Number of particles to generate.
    rx, ry, rzeta : float
        Scaling factors for the x, px, y, py, zeta, and pzeta dimensions respectively.
    rng_seed : int
        Seed for the random number generator for reproducibility.

    Returns
    -------
    x_norm, px_norm, y_norm, py_norm, zeta_norm, pzeta_norm : np.ndarray
        Coordinates of the generated points in the 6-dimensional space.
    '''

    x_norm , px_norm , y_norm, py_norm, zeta_norm, pzeta_norm = generate_hypersphere(num_particles,D=6,r=[rx,rx,ry,ry,rzeta,rzeta], rng_seed=rng_seed, surface = False,unpack=True)

    return x_norm , px_norm , y_norm, py_norm, zeta_norm, pzeta_norm


# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import numpy as np

import xobjects as xo
import xpart as xp
import xtrack as xt

from xobjects.test_helpers import for_all_test_contexts, fix_random_seed

test_data_folder = xt._pkg_root.joinpath('../test_data').absolute()

@for_all_test_contexts
@fix_random_seed(6453645)
def test_multi_bunch_gaussian_generation(test_context):
    bunch_intensity = 1e11
    sigma_z = 22.5e-2 / 5
    n_part_per_bunch = int(1e5)
    nemitt_x = 2e-6
    nemitt_y = 2.5e-6

    filename = test_data_folder.joinpath(
        'lhc_no_bb/line_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
    line = xt.Line.from_dict(input_data['line'])
    line.particle_ref = xp.Particles.from_dict(
        input_data['particle'],
        _context=test_context  # for testing purposes
    )

    line.build_tracker(_context=test_context)

    part_on_co = line.find_closed_orbit()
    tw = line.twiss()
    sigma_delta = sigma_z / tw['bets0']
    circumference = line.get_length()
    h_list = [35640]
    bunch_spacing_in_buckets = 10
    bucket_length = circumference/h_list[0]
    bunch_spacing = bunch_spacing_in_buckets * bucket_length
    filling_scheme = np.zeros(int(np.amin(h_list)/bunch_spacing_in_buckets))
    # build a dummy filling scheme
    n_bunches_tot = 10
    filling_scheme[0:int(n_bunches_tot/2)] = 1
    filling_scheme[n_bunches_tot:int(3*n_bunches_tot/2)] = 1
    filled_slots = filling_scheme.nonzero()[0]

    # make a test faking 3 procs sharing the bunches
    n_procs = 3

    bunch_selection_per_rank = xp.matched_gaussian.split_scheme(filling_scheme=filling_scheme,
                                            n_chunk=n_procs)


    expected_bunch_selection_per_rank = {
        0: np.array([0, 1, 2, 3]),
        1: np.array([4, 5, 6]),
        2: np.array([7, 8, 9])
    }

    for rank in range(n_procs):
        xo.assert_allclose(bunch_selection_per_rank[rank],
                           expected_bunch_selection_per_rank[rank],
                           atol=0, rtol=1e-15)
        part = xp.generate_matched_gaussian_multibunch_beam(
            _context=test_context,
            filling_scheme=filling_scheme,
            bunch_num_particles=n_part_per_bunch,
            bunch_intensity_particles=bunch_intensity,
            nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
            line=line, bunch_spacing_buckets=bunch_spacing_in_buckets,
            bunch_selection=bunch_selection_per_rank[rank],
            particle_ref=line.particle_ref
        )

        assert len(part.x) == n_part_per_bunch*len(bunch_selection_per_rank[rank])

        for i_bunch,bunch_number in enumerate(bunch_selection_per_rank[rank]):
            zeta_avg = np.average(
            test_context.nparray_from_context_array(
                part.zeta[i_bunch*n_part_per_bunch:
                       (i_bunch+1)*n_part_per_bunch]))
            delta_avg = np.average(
            test_context.nparray_from_context_array(
                part.delta[i_bunch*n_part_per_bunch:
                       (i_bunch+1)*n_part_per_bunch]))
            y_rms = np.std(
                test_context.nparray_from_context_array(
                    part.y[i_bunch*n_part_per_bunch:
                           (i_bunch+1)*n_part_per_bunch]))
            x_rms = np.std(
                test_context.nparray_from_context_array(
                    part.x[i_bunch*n_part_per_bunch:
                           (i_bunch+1)*n_part_per_bunch]))
            delta_rms = np.std(
                test_context.nparray_from_context_array(
                    part.delta[i_bunch*n_part_per_bunch:
                               (i_bunch+1)*n_part_per_bunch]))
            zeta_rms = np.std(
                test_context.nparray_from_context_array(
                    part.zeta[i_bunch*n_part_per_bunch:
                              (i_bunch+1)*n_part_per_bunch]))

            assert np.isclose((zeta_avg+bunch_spacing*filled_slots[bunch_number])/sigma_z, 0.0, atol=1e-2)
            assert np.isclose(delta_avg/sigma_delta, 0.0, atol=1e-2)
            assert np.isclose(zeta_rms, sigma_z, rtol=1e-2, atol=1e-15)
            assert np.isclose(delta_rms, sigma_delta, rtol=1e-1, atol=1e-15)

            part_on_co.move(_context=xo.ContextCpu(omp_num_threads=0))

            gemitt_x = nemitt_x/part_on_co.beta0/part_on_co.gamma0
            gemitt_y = nemitt_y/part_on_co.beta0/part_on_co.gamma0
            assert np.isclose(
                x_rms,
                np.sqrt(tw['betx'][0]*gemitt_x + tw['dx'][0]**2*delta_rms**2),
                rtol=1e-2, atol=1e-15
            )
            assert np.isclose(
                y_rms,
                np.sqrt(tw['bety'][0]*gemitt_y + tw['dy'][0]**2*delta_rms**2),
                rtol=1e-2, atol=1e-15
            )

            assert np.isclose(
                np.sum(test_context.nparray_from_context_array(
                part.weight[i_bunch*n_part_per_bunch:
                       (i_bunch+1)*n_part_per_bunch])),
                bunch_intensity,
                rtol=1e-2, atol=1e-15
            )




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

'''
This module was originally written for PyHEADTAIL
@author Kevin Li, Adrian Oeftiger, Stefan Hegglin
@date 16.06.2017
@brief module for matching longitudinal particle distributions to an RFBucket instance
'''
import numpy as np
import logging

from scipy.optimize import brentq, newton
from scipy.integrate import fixed_quad
from scipy.constants import e, c
from functools import partial
from abc import abstractmethod

from . import pdf_integrators_2d as integr
from ..general import _print

logger = logging.getLogger(__name__)


class RFBucketMatcher:

    integrationmethod = ['quad', 'cumtrapz'][0]
    def get_moment_integrators(self):
        '''Return moment integrators from
        cobra_functions.pdf_integrators_2d according to the chosen
        self.integrationmethod. Allows to change integration method
        for RFBucket matching.
        '''
        zero = getattr(integr, 'compute_zero_' + self.integrationmethod)
        mean = getattr(integr, 'compute_mean_' + self.integrationmethod)
        var = getattr(integr, 'compute_var_' + self.integrationmethod)
        cov = getattr(integr, 'compute_cov_' + self.integrationmethod)
        return zero, mean, var, cov

    def __init__(self, rfbucket, distribution_type=None, sigma_z=None,
                 epsn_z=None, verbose_regeneration=False, psi=None,
                 *args, **kwargs):

        if psi is not None:
            raise NotImplemented
            # _print('Warning! Keyword argument "psi" will be removed in a future '
            #               'PyHEADTAIL release!')
            # if distribution_type is not None:
            #     raise TypeError('RFBucketMatcher accepts either '
            #                     'distribution_type or psi as argument. ')
        else:
            if distribution_type is None:
                raise TypeError(
                    '__init__() takes at least 2 arguments (1 given)\n'
                    '--> kwarg psi has been renamed to distribution_type.')

        self.rfbucket = rfbucket
        hamiltonian = partial(rfbucket.hamiltonian, make_convex=True)
        hmax = rfbucket.h_sfp(make_convex=True)

        self.psi_object = distribution_type(hamiltonian, hmax)
        self.psi = self.psi_object.function

        self.verbose_regeneration = verbose_regeneration

        if sigma_z and not epsn_z:
            self.variable = sigma_z
            self.psi_for_variable = self.psi_for_bunchlength_newton_method
        elif not sigma_z and epsn_z:
            self.variable = epsn_z
            self.psi_for_variable = self.psi_for_emittance_newton_method
        else:
            raise ValueError("Can not generate mismatched matched "
                             "distribution! (Don't provide both sigma_z "
                             "and epsn_z!)")

    ### analytic matching methods:

    def psi_for_emittance_newton_method(self, epsn_z):
        # Maximum emittance
        self.psi_object.H0 = self.rfbucket.guess_H0(
            self.rfbucket.circumference, from_variable='sigma')
        epsn_max = self._compute_emittance(self.rfbucket, self.psi)
        if epsn_z > epsn_max:
            logger.warning('Given RMS emittance does not fit into bucket. '
                           'Using (maximum) full bucket emittance ' +
                           str(epsn_max*0.99) + 'eV s instead.')
            epsn_z = epsn_max*0.99
        _print('*** Maximum RMS emittance ' + str(epsn_max) + 'eV s.')

        def error_from_target_epsn(ec):
            self.psi_object.H0 = self.rfbucket.guess_H0(
                ec, from_variable='epsn')
            emittance = self._compute_emittance(self.rfbucket, self.psi)

            if np.isnan(emittance): raise ValueError

            _print('... distance to target emittance: ' +
                        '{:.2e}'.format(emittance-epsn_z))

            return emittance-epsn_z

        try:
            ec_bar = brentq(error_from_target_epsn, epsn_z/100, 2*epsn_max,
                            rtol=1e-5)
        except ValueError:
            logger.warning(
                'RFBucketMatcher: failed to converge with Brent method, '
                'continuing with Newton-Raphson method.')
            ec_bar = newton(error_from_target_epsn, epsn_z, tol=1e-5)

        self.psi_object.H0 = self.rfbucket.guess_H0(
            ec_bar, from_variable='epsn')
        emittance = self._compute_emittance(self.rfbucket, self.psi)
        _print('--> Emittance: ' + str(emittance))
        sigma = self._compute_sigma(self.rfbucket, self.psi)
        _print('--> Bunch length: ' + str(sigma))

    def psi_for_bunchlength_newton_method(self, sigma):
        # Maximum bunch length
        self.psi_object.H0 = self.rfbucket.guess_H0(
            self.rfbucket.circumference, from_variable='sigma')
        sigma_max = self._compute_sigma(self.rfbucket, self.psi)
        if sigma > sigma_max:
            logger.warning('Given RMS bunch length does not fit into bucket. '
                           'Using (maximum) full bucket RMS bunch length ' +
                           str(sigma_max*0.99) + 'm instead.')
            sigma = sigma_max*0.99
        _print('*** Maximum RMS bunch length ' + str(sigma_max) + 'm.')

        def error_from_target_sigma(sc):
            '''Width for bunch length'''
            self.psi_object.H0 = self.rfbucket.guess_H0(
                sc, from_variable='sigma')
            length = self._compute_sigma(self.rfbucket, self.psi)

            if np.isnan(length): raise ValueError

            _print('... distance to target bunch length: ' +
                        '{:.4e}'.format(length-sigma))

            return length-sigma

        try:
            sc_bar = brentq(error_from_target_sigma, sigma/100, 2*sigma_max,
                            rtol=1e-5)
        except ValueError:
            logger.warning(
                'RFBucketMatcher: failed to converge with Brent method, '
                'continuing with Newton-Raphson method.')
            sc_bar = newton(error_from_target_sigma, sigma, tol=1e-5)

        self.psi_object.H0 = self.rfbucket.guess_H0(
            sc_bar, from_variable='sigma')
        sigma = self._compute_sigma(self.rfbucket, self.psi)
        _print('--> Bunch length: ' + str(sigma))
        emittance = self._compute_emittance(self.rfbucket, self.psi)
        _print('--> Emittance: ' + str(emittance))

    def linedensity(self, xx, quad_type=fixed_quad):
        L = []
        try:
            L = np.array([quad_type(lambda y: self.psi(x, y), 0,
                                    self.rfbucket.separatrix(x))[0]
                          for x in xx])
        except TypeError:
            L = quad_type(lambda y: self.psi(xx, y), 0,
                          self.rfbucket.separatrix(xx))[0]
        L = np.array(L)

        return 2*L

    def generate(self, macroparticlenumber, cutting_margin=0):
        '''Generate a 2d phase space of n_particles particles randomly distributed
        according to the particle distribution function psi within the region
        [xmin, xmax, ymin, ymax].
        '''
        self.psi_for_variable(self.variable)

        xmin, xmax = self.rfbucket.z_left, self.rfbucket.z_right
        ymin = -self.rfbucket.dp_max(self.rfbucket.z_right)
        ymax = -ymin

        # rejection sampling
        uniform = np.random.uniform
        n_gen = macroparticlenumber
        u = uniform(low=xmin, high=xmax, size=n_gen)
        v = uniform(low=ymin, high=ymax, size=n_gen)
        s = uniform(size=n_gen)

        def mask_out(s, u, v):
            return s >= self.psi(u, v)

        if cutting_margin:
            mask_out_nocut = mask_out

            def mask_out(s, u, v):
                return np.logical_or(
                    mask_out_nocut(s, u, v),
                    ~self.rfbucket.is_in_separatrix(u, v, cutting_margin))

        # masked_out = ~(s<self.psi(u, v))
        masked_out = mask_out(s, u, v)
        while np.any(masked_out):
            masked_ids = np.where(masked_out)[0]
            n_gen = len(masked_ids)
            u[masked_out] = uniform(low=xmin, high=xmax, size=n_gen)
            v[masked_out] = uniform(low=ymin, high=ymax, size=n_gen)
            s[masked_out] = uniform(size=n_gen)
            # masked_out = ~(s<self.psi(u, v))
            masked_out[masked_ids] = mask_out(
                s[masked_out], u[masked_out], v[masked_out]
            )
            if self.verbose_regeneration:
                _print(
                    'Thou shalt not give up! :-) '
                    'Regenerating {0} macro-particles...'.format(n_gen))

        return u, v, self.psi, self.linedensity

    def _compute_sigma(self, rfbucket, psi):
        z_left = rfbucket.z_left
        z_right = rfbucket.z_right
        zero, mean, var, cov = self.get_moment_integrators()

        var_x = var(self.psi, lambda x: -rfbucket.separatrix(x),
                    rfbucket.separatrix, z_left, z_right,
                    direction='x') # x means z direction

        return np.sqrt(var_x)

    def _compute_emittance(self, rfbucket, psi):
        z_left = rfbucket.z_left
        z_right = rfbucket.z_right
        zero, mean, var, cov = self.get_moment_integrators()

        var_x, cov_xy, var_y = cov(
            self.psi, lambda x: -rfbucket.separatrix(x),
            rfbucket.separatrix, z_left, z_right)

        return (np.sqrt(var_x*var_y - cov_xy**2) *
                4*np.pi*rfbucket.p0/np.abs(rfbucket.charge_coulomb))


class StationaryDistribution(object):
    def __init__(self, H, Hmax=None, Hcut=0, H0=1):
        self.H = H
        self.H0 = H0
        if not Hmax:
            self.Hmax = H(0, 0)
        else:
            self.Hmax = Hmax
        self.Hcut = Hcut

    @abstractmethod
    def _psi(self, H):
        '''Define the distribution value for the given H, the output
        lies in the interval [0,1]. This is the central function to
        be implemented by stationary distributions.
        '''
        pass

    def function(self, z, dp):
        psi = self._psi(self.H(z, dp).clip(min=self.Hcut))
        norm = self._psi(self.Hmax)
        return psi / norm

class ThermalDistribution(StationaryDistribution):
    '''Thermal Boltzmann distribution psi ~ exp(-H/H0).
    For a quadratic harmonic oscillator Hamiltonian this gives the
    bi-Gaussian phase space distribution.
    '''
    def _psi(self, H):
        # convert from convex Hamiltonian
        # (SFP being the maximum and the separatrix having zero value)
        # to conventional literature scale (zero-valued minimum at SFP)
        Hsep = self.Hcut + self.Hmax
        Hn = Hsep - H
        # if self.H0<1:
        #     print('Hn', Hn)
        #     print('H0', self.H0)
        #     breakpoint()
        # f(Hn) - f(Hsep)
        return np.exp(-Hn / self.H0) - np.exp(-Hsep / self.H0)

class QGaussianDistribution(StationaryDistribution):
    '''Specific Tsallis q-Gaussian distribution for q=3/5 for now,
    leading to psi ~ (1 - H/H0)^2, this may be generalised.
    '''
    n = 2

    def _psi(self, H):
        # convert from convex Hamiltonian
        # (SFP being the maximum and the separatrix having zero value)
        # to conventional scale (zero-valued minimum at SFP)
        Hsep = self.Hcut + self.Hmax
        Hn = Hsep - H
        dist = (1 - Hn / self.H0).clip(min=self.Hcut)**self.n
        return dist.clip(min=self.Hcut)

class ParabolicDistribution(StationaryDistribution):
    '''The parabolic profile distribution is a specific case of the
    present implementation of the q-Gaussian distribution for n = 1/2,
    psi ~ sqrt(1 - H/H0).
    For a quadratic harmonic oscillator Hamiltonian this distribution
    provides a parabolic line density.
    '''
    def _psi(self, H):
        # convert from convex Hamiltonian
        # (SFP being the maximum and the separatrix having zero value)
        # to conventional scale (zero-valued minimum at SFP)
        Hsep = self.Hcut + self.Hmax
        Hn = Hsep - H
        return np.sqrt((1 - Hn / self.H0).clip(min=self.Hcut))

class WaterbagDistribution(StationaryDistribution):
    '''The waterbag distribution has a constant Hamiltonian distribution
    until a cutoff, psi ~ Theta(H - H0) with Theta the Heaviside
    step function.
    '''
    def _psi(self, H):
        # convert from convex Hamiltonian
        # (SFP being the maximum and the separatrix having zero value)
        # to conventional scale (zero-valued minimum at SFP)
        Hsep = self.Hcut + self.Hmax
        Hn = Hsep - H
        return np.piecewise(Hn, [Hn <= self.H0, Hn > self.H0], [1./self.H0, 0])



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import json

import xpart as xp
import xtrack as xt

from xobjects.test_helpers import for_all_test_contexts

test_data_folder = xt._pkg_root.joinpath('../test_data').absolute()


@for_all_test_contexts
def test_classical_particle_radius_ions(test_context):
    """
    Test classical particle radius for ions
    """
    # Load SPS ion sequence with Pb82 ions
    fname_line = test_data_folder.joinpath('sps_ions/line_and_particle.json')
    with open(fname_line, 'r') as fid:
        input_data = json.load(fid)

    particle_ref = xp.Particles.from_dict(input_data['particle_ref'],
                                          _context=test_context)
    r0_Pb82 = np.float64(4.998617e-17)  # calculated theoretical classical particle radius

    assert np.isclose(particle_ref.get_classical_particle_radius0(), r0_Pb82, atol=1e-5)


@for_all_test_contexts
def test_classical_particle_radius_protons(test_context):
    """
    Test classical particle radius for protons
    """
    # Load SPS sequence without space charge
    fname_line = test_data_folder.joinpath('sps_w_spacecharge/'
                                           'line_no_spacecharge_and_particle.json')
    with open(fname_line, 'r') as fid:
        input_data = json.load(fid)

    particle_ref = xp.Particles.from_dict(input_data['particle'],
                                          _context=test_context)
    r0_proton = np.float64(1.534698e-18)  # calculated theoretical classical particle radius

    assert np.isclose(particle_ref.get_classical_particle_radius0(), r0_proton, atol=1e-5)


def test_showing():
    """
    Test whether showing particle values gives reasonable output
    """
    test_data_folder = xt._pkg_root.joinpath('../test_data').absolute()

    # Load SPS ion sequence with Pb82 ions
    fname_line = test_data_folder.joinpath('sps_ions/line_and_particle.json')
    with open(fname_line, 'r') as fid:
        input_data = json.load(fid)

    particle_ref = xp.Particles.from_dict(input_data['particle_ref'])
    particle_ref.show()



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
import numpy as np

import xpart as xp
import xtrack as xt

from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_pencil(test_context):
    num_particles = 10000
    nemitt_x = 2.5e-6
    nemitt_y = 3e-6

    # Load machine model (from pymask)
    filename = xt._pkg_root.parent.joinpath('test_data/lhc_no_bb/line_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
    line = xt.Line.from_dict(input_data['line'])
    line.build_tracker()
    particle_sample = xp.Particles.from_dict(input_data['particle'])

    # Horizontal plane: generate gaussian distribution in normalized coordinates
    x_in_sigmas, px_in_sigmas = xp.generate_2D_gaussian(num_particles)

    # Vertical plane: generate pencil distribution in normalized coordinates
    pencil_cut_sigmas = 6.
    pencil_dr_sigmas = 0.7
    y_in_sigmas, py_in_sigmas, r_points, theta_points = xp.generate_2D_pencil(
                                 num_particles=num_particles,
                                 pos_cut_sigmas=pencil_cut_sigmas,
                                 dr_sigmas=pencil_dr_sigmas,
                                 side='+-')

    # Longitudinal plane: generate gaussian distribution matched to bucket
    zeta, delta = xp.generate_longitudinal_coordinates(
            num_particles=num_particles, distribution='gaussian',
            sigma_z=10e-2, particle_ref=particle_sample, line=line)

    # Build particles:
    #    - scale with given emittances
    #    - transform to physical coordinates (using 1-turn matrix)
    #    - handle dispersion
    #    - center around the closed orbit
    particles = xp.build_particles(_context=test_context,
                                   line=line,
                                   particle_ref=particle_sample,
                                   zeta=zeta, delta=delta,
                                   x_norm=x_in_sigmas, px_norm=px_in_sigmas,
                                   y_norm=y_in_sigmas, py_norm=py_in_sigmas,
                                   nemitt_x=nemitt_x, nemitt_y=nemitt_y)

    dct = particles.to_dict() # transfers it to cpu
    assert np.min(np.abs(dct['y'])) > 0.0018
    assert np.max(np.abs(dct['y'])) < 0.0021



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
import pytest

import xpart as xp

from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_to_from_dict(test_context):
    # Create a Particles on your selected context (default is CPU)
    part = xp.Particles(_context=test_context, x=[1, 2, 3])

    # Save particles to dict
    dct = part.to_dict()

    # Load particles from dict
    part_from_dict = xp.Particles.from_dict(dct, _context=test_context)

    #!end-doc-part
    dct = part.to_dict()
    for nn in 'x px y py zeta delta ptau rpp rvv gamma0 p0c'.split():
        assert isinstance(dct[nn], np.ndarray)
        assert isinstance(getattr(part_from_dict, nn), test_context.nplike_array_type)


def test_to_dict_pyheadtail_interface():
    xp.enable_pyheadtail_interface()
    assert xp.Particles.__name__ == 'PyHtXtParticles'
    xp.disable_pyheadtail_interface()
    assert xp.Particles.__name__ == 'Particles'

    p = xp.pyheadtail_interface.pyhtxtparticles.PyHtXtParticles(x=[1, 2, 3])

    dct = p.to_dict()
    p1 = xp.Particles.from_dict(dct)

    assert p1.x[1] == 2


@for_all_test_contexts
@pytest.mark.parametrize('compact', [False, True])
def test_to_pandas(test_context, compact):
    n_particles = 1000
    capacity = 3000
    part = xp.Particles(_capacity=capacity, _context=test_context,
                x=np.random.uniform(low=-1e-6, high=1e-6, size=n_particles),
                px=np.random.uniform(low=-1e-6, high=1e-6, size=n_particles),
                y=np.random.uniform(low=-1e-3, high=1e-3, size=n_particles),
                py=np.random.uniform(low=-1e-6, high=1e-6, size=n_particles),
                zeta=np.random.uniform(low=-1e-2, high=1e-2, size=n_particles),
                delta=np.random.uniform(low=-1e-4, high=1e-4, size=n_particles),
                p0c=7e12)

    df_part = part.to_pandas(compact=compact)

    if compact:
        assert len(df_part) == n_particles
    else:
        assert len(df_part) == capacity

    # Check that underscored vars (if any) are all rng states
    ltest = [nn for nn in df_part.keys() if nn.startswith('_')]
    assert np.all([nn.startswith('_rng') for nn in ltest])

    for nn in ['_rng_s1', '_rng_s2', '_rng_s3', '_rng_s4',
                'beta0', 'gamma0', 'ptau', 'rpp', 'rvv']:
        if compact:
            assert nn not in df_part.keys()
        else:
            assert nn in df_part.keys()

    if compact:
        part = part.remove_unused_space()

    part_test = xp.Particles.from_pandas(df_part)
    for kk in ['x', 'px', 'y', 'py', 'zeta', 'delta', 'ptau', 'gamma0']:
        assert np.all(part.to_dict()[kk] == part_test.to_dict()[kk])


# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xpart as xp

from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_basics(test_context):
    p1 = xp.Particles(x=[1, 2, 3], px=[10, 20, 30],
                      mass0=xp.ELECTRON_MASS_EV,
                      _context=test_context)

    mask = p1.x > 1

    p2 = p1.filter(mask)

    assert p2._buffer.context == test_context
    assert p2._capacity == 2
    dct = p2.to_dict()
    assert dct['mass0'] == xp.ELECTRON_MASS_EV
    assert np.all(dct['px'] == np.array([20., 30.]))



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xpart as xp
import xtrack as xt

from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_to_from_dict_longitudinal_consistency(test_context):
    cav = xt.Cavity(_context=test_context, frequency=400e6, voltage=6e6)

    part = xp.Particles(_context=test_context, p0c=6500e9, x=[1, 2, 3], delta=1e-4)
    cav.track(part)

    part2 = xp.Particles(_context=test_context, **part.to_dict())

    tocpu = test_context.nparray_from_context_array
    for nn in 'x px y py zeta delta ptau rpp rvv gamma0 p0c'.split():
        assert np.all(tocpu(getattr(part, nn)) == tocpu(getattr(part2, nn)))


@for_all_test_contexts
def test_to_from_pandas_longitudinal_consistency(test_context):
    cav = xt.Cavity(_context=test_context, frequency=400e6, voltage=6e6)

    part = xp.Particles(_context=test_context, p0c=6500e9, x=[1, 2, 3], delta=1e-4)
    cav.track(part)

    df = part.to_pandas()
    part2 = xp.Particles.from_pandas(df, _context=test_context)

    tocpu = test_context.nparray_from_context_array
    for nn in 'x px y py zeta delta ptau rpp rvv gamma0 p0c'.split():
        assert np.all(tocpu(getattr(part, nn)) == tocpu(getattr(part2, nn)))



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
from scipy.constants import c
import scipy.special
import scipy.integrate as integrate
from scipy.special import gamma as Gamma

from ..general import _print


class SingleRFHarmonicMatcher:
    def __init__(self,
                 q0=None,
                 voltage=None,
                 length=None,
                 freq=None,
                 p0c=None,
                 slip_factor=None,
                 beta0=None,
                 rms_bunch_length=None, distribution="parabolic",
                 transformation_particles=400000, n_points_in_distribution=300,
                 verbose=0, m=4.7, q=1.0):

        self.verbose = verbose
        self.transformation_particles = transformation_particles

        self.length = length

        # Hamoltonian: H = A cos(B tau) - C ptau^2
        # normalized Hamiltonian: m = ( sin(B/2*tau) )^2 + C/(2A) ptau^ 2
        self.A = q0*voltage/(2.*np.pi*freq*p0c/c*length)
        self.B = 2*np.pi*freq/c
        self.C = abs(slip_factor)/(2.*beta0*beta0)
        # the difference between above and below transition is that the Hamiltonian flips sign
        # (considering always the absolute value of the slip factor)a. This is the same as if
        # the particle goes back in turns. For the purpose of matching, this is of no concern.

        tau_ufp = self.get_unstable_fixed_point()
        dtau = 0.01*tau_ufp # don't get too close to the separatrix
        tau_lim = tau_ufp - dtau
        self.tau_distr_x = np.linspace(- tau_lim, tau_lim, n_points_in_distribution)
        if distribution == "parabolic":
            tau_max = np.sqrt(5)*rms_bunch_length
            lambda_dist = lambda tau, tau_max: 1 - (tau/tau_max)**2
            # correct bunch length (due to truncation at separatrix)
            if tau_max > tau_lim:
                # analytical_tau_max = np.sqrt((5*tau_lim**3*rms_bunch_length**2 - 3*tau_lim**5)/(15*tau_lim*rms_bunch_length**2 - 5*tau_lim**3))
                func_to_solve = lambda new_tau_max: (scipy.integrate.quad(lambda x: (x**2 - rms_bunch_length**2)*lambda_dist(x, new_tau_max), -tau_lim, tau_lim))[0]
                corrected_tau_max = scipy.optimize.fsolve(func_to_solve, x0=tau_max)[0]
                tau_max = corrected_tau_max
            self.tau_distr_y = lambda_dist(self.tau_distr_x, tau_max)
            self.tau_distr_y[abs(self.tau_distr_x) > tau_max] = 0
            if tau_max >= tau_ufp:
                _print(f"WARNING SingleRFHarmonicMatcher: longitudinal profile larger than bucket, truncating to unstable fixed point. tau_max = {tau_max:.4f}, tau_ufp = {tau_ufp:.4f} ")

            _print(f"SingleRFHarmonicMatcher: Parabolic parameter is equal to {tau_max:.3f}m to achieve target RMS bunch length ({rms_bunch_length:.3f}m).")
        elif distribution == "gaussian":
            lambda_dist = lambda tau, rms: np.exp(-tau**2/2./rms**2)

            # correct bunch length (due to truncation at separatrix)
            func_to_solve = lambda new_rms: (scipy.integrate.quad(lambda x: (x**2 - rms_bunch_length**2)*lambda_dist(x, new_rms), -tau_lim, tau_lim))[0]
            corrected_rms = scipy.optimize.fsolve(func_to_solve, x0=rms_bunch_length)[0]

            _print(f"SingleRFHarmonicMatcher: Gaussian parameter is equal to {corrected_rms:.3f}m to achieve target RMS bunch length ({rms_bunch_length:.3f}m).")

            self.tau_distr_y = lambda_dist(self.tau_distr_x, corrected_rms)
        elif distribution == "qgaussian":
            # Q-Gaussian from Eq. (2.1) in (Umarov, Tsallis, Steinberg, 2008) 
            #available at https://link.springer.com/article/10.1007/s00032-008-0087-y
            if q>5./3.:
                q = 1.6 # slightly below defined region
                _print(f"WARNING SingleRFHarmonicMatcher: q-value above 5/3 undefined for correct RMS bunch length, truncating q value to {q:.3f}")
            beta = 1.0/(rms_bunch_length**2 * (5.-3.*q)) # solving from variance
            lambda_dist = lambda tau, beta: np.sqrt(beta) / self._Cq(q) * self._eq(-beta * tau**2, q)
            
            func_to_solve = lambda new_beta: (scipy.integrate.quad(lambda x: (x**2 - rms_bunch_length**2)*lambda_dist(x, new_beta), -tau_lim, tau_lim))[0]
            corrected_beta = scipy.optimize.fsolve(func_to_solve, x0=beta)[0]
            _print(f"SingleRFHarmonicMatcher: q-Gaussian parameter beta = {corrected_beta:.3f} for q={q:.3f} to achieve target RMS bunch length ({rms_bunch_length:.3f}m).")

            self.tau_distr_y = lambda_dist(self.tau_distr_x, corrected_beta)
            
        elif distribution == "binomial":
            # Binomial distribution adds tail to parabolic, see (Joho, 1980) at https://indico.psi.ch/event/3484/attachments/5948/7502/TM-11-14.pdf
            # behaviour is Gaussian for m --> inf
            tau_max = 1.0 # starting value, will be adjusted. Used as benchmarking value with RMS factor for parabola
            lambda_dist = lambda tau, tau_max: (1 - (tau/tau_max)**2)**(m-0.5) 
            binomial_2nd = lambda tau, tau_max: (1 - (tau/tau_max)**2)**(m-0.5)*tau**2
            RMS_binomial = np.sqrt(integrate.quad(binomial_2nd, -1, 1, args=(tau_max))[0] / integrate.quad(lambda_dist, -1, 1, args=(tau_max))[0])
            factor_binomial = tau_max / RMS_binomial
            _print(f"RMS factor for binimial is {factor_binomial:.3f}")
            tau_max = factor_binomial*rms_bunch_length 
            func_to_solve = lambda new_tau_max: (scipy.integrate.quad(lambda x: (x**2 - rms_bunch_length**2)*lambda_dist(x, new_tau_max), -tau_lim, tau_lim))[0]
            corrected_tau_max = scipy.optimize.fsolve(func_to_solve, x0=tau_max)[0]
            tau_max = corrected_tau_max
            self.tau_distr_y = lambda_dist(self.tau_distr_x, tau_max)
            self.tau_distr_y[abs(self.tau_distr_x) > tau_max] = 0
            _print(f"SingleRFHarmonicMatcher: Binomial x_lim parameter is equal to {tau_max:.3f}m to achieve target RMS bunch length ({rms_bunch_length:.3f}m).")
        else:
            raise NotImplementedError

        self.m_distr_x, self.m_distr_y = self.transform_tau_distr_to_m_distr()


    def transform_tau_distr_to_m_distr(self):
        xp = self.tau_distr_x.copy()
        yp = self.tau_distr_y.copy()
        N = int(len(xp)/2.)
        dx = xp[1] - xp[0]
        m_distr_x = []
        m_distr_y = []
        for ii in range(N):
            jj = len(xp) - ii - 1 #start from end
            tau = xp[jj]
            dens = yp[jj]
            if dens == 0.:
                continue
            if self.verbose:
                _print(f"tau = {tau:.3f}, f(tau) = {dens:.3f}")

            m0 = self.get_m(tau=tau - dx/2.)
            m1 = self.get_m(tau=tau + dx/2.)
            dm = m1 - m0
            m = self.get_m(tau=tau)
            if m == 1.0:
                continue
            _print('SingleRFHarmonicMatcher: Transforming distribution: '
                        f'{round(ii/N*100):2d}%  ',end="\r", flush=True)
            tau_test, ptau_test = self.get_airbag_from_m(m=m, n_particles=self.transformation_particles)
            hist, bin_edges = np.histogram(tau_test, bins=len(xp), range=(min(xp)-dx/2., max(xp)+dx/2.))

            hist = (hist + hist[::-1])/2.
            factor = dens/hist[jj]
            yp -= hist*factor
            m_distr_x.append(m)
            m_distr_y.append(np.sum(hist)*factor/dm)

        m_distr_x.append(0)
        m_distr_y.append(0)
        m_distr_x = m_distr_x[::-1]
        m_distr_y = m_distr_y[::-1]

        _print('SingleRFHarmonicMatcher: Done transforming distribution.')
        return m_distr_x, m_distr_y


    def get_separatrix(self):
        ufp = self.get_unstable_fixed_point()
        xx = np.linspace(-ufp, ufp, 1000)
        yy = np.sqrt(2*self.A/self.C) * np.cos(self.B/2.*xx)
        return xx, yy

    def get_unstable_fixed_point(self):
        return np.pi/self.B

    def get_m(self, tau=0, ptau=0):
        return ( np.sin(self.B/2. * tau) )**2 + self.C / 2. / self.A * (ptau ** 2)

    def get_airbag_from_m(self, m, n_particles=20000):
        if n_particles is None:
            n_particles = len(m)

        K = scipy.special.ellipk(m)
        G = 2.*K/np.pi
        theta = np.random.uniform(size=n_particles)*2.*np.pi
        sn, cn, dn, ph = scipy.special.ellipj(G*theta,m)

        tau = 2./self.B*np.arcsin(np.sqrt(m)*sn)
        ptau = np.sqrt(2*self.A*m/self.C)*cn

        return tau, ptau

    def sample_tau_ptau(self, n_particles=20000):
        max_m = max(self.m_distr_x)
        tau_new = []
        ptau_new = []
        counter = 0
        max_y = np.max(self.m_distr_y)
        chunk = 20000
        ### Acceptance-rejection algorithm sampling from distribution of m and a random "angle"
        ### The random angle is the conjugate variable to the action variable and is only
        ### approximately equal to the angle in the tau-ptau space.
        while counter < n_particles:
            m = np.random.random(size=chunk)*max_m
            rand_test = np.random.random(size=chunk)*max_y
            yy = np.interp(m, self.m_distr_x, self.m_distr_y)

            tau, ptau = self.get_airbag_from_m(m, n_particles=None)

            mask = rand_test < yy

            tau_new.extend(list(tau[mask]))
            ptau_new.extend(list(ptau[mask]))
            counter += sum(mask)
            _print('SingleRFHarmonicMatcher: Sampling particles: '
                        f'{round(counter/n_particles*100):2d}%  ',end="\r", flush=True)
        _print(f"SingleRFHarmonicMatcher: Sampled {n_particles} particles")

        return tau_new[:n_particles], ptau_new[:n_particles]

    def generate(self, n_particles=20000):
        tau, ptau = self.sample_tau_ptau(n_particles=n_particles)

    def get_synchrotron_tune(self):
        return self.B*np.sqrt(2*self.A*self.C)*self.length/(2*np.pi)

    def _Cq(self, q, margin=5e-4):
        """
        Normalization coefficient for Q-Gaussian
        
        Normalizing constant from Eq. (2.2) in https://link.springer.com/article/10.1007/s00032-008-0087-y
        with a small margin around 1.0 for numerical stability
        """
        if q < (1 - margin):
            Cq = (2 * np.sqrt(np.pi) * Gamma(1.0/(1.0-q))) / ((3.0 - q) * np.sqrt(1.0 - q) * Gamma( (3.0-q)/(2*(1.0 -q))))   
        elif (q > (1.0 - margin) and q < (1.0 + margin)):
            Cq = np.sqrt(np.pi)
        else:
            Cq = (np.sqrt(np.pi) * Gamma((3.0-q)/(2*(q-1.0)))) / (np.sqrt(q-1.0) * Gamma(1.0/(q-1.0)))
        if q > 3.0:
            raise ValueError("q must be smaller than 3!")
        else:
            return Cq
    
    
    def _eq(self, x, q):
        """ 
        Q-exponential function
        Available at https://link.springer.com/article/10.1007/s00032-008-0087-y
        """
        eq = np.zeros(len(x))
        for i, xx in enumerate(x):
            if ((q != 1) and (1 + (1 - q) * xx) > 0):
                eq[i] = (1 + (1 - q) * xx)**(1 / (1 - q))
            elif q==1:
                eq[i] = np.exp(xx)
            else:
                eq[i] = 0
        return eq



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

'''
This module was originally written for PyHEADTAIL
@author Kevin Li, Adrian Oeftiger
'''

import numpy as np
from scipy.constants import c
from scipy.optimize import newton
from scipy.integrate import dblquad
from functools import partial, wraps

from .curve_tools import zero_crossings as cvt_zero_crossings

from functools import reduce


def attach_clean_buckets(rf_parameter_changing_method, rfsystems_instance):
    '''Wrap an rf_parameter_changing_method (that changes relevant RF
    parameters, i.e. Kick attributes). Needs to be an instance method,
    presumably an RFSystems instance (hence the self argument in
    cleaned_rf_parameter_changing_method).
    In detail, attaches a call to the rfsystems_instance.clean_buckets
    method after calling the wrapped function.
    '''
    @wraps(rf_parameter_changing_method)
    def cleaned_rf_parameter_changing_method(self, *args, **kwargs):
        res = rf_parameter_changing_method(*args, **kwargs)
        rfsystems_instance.clean_buckets()
        return res
    return cleaned_rf_parameter_changing_method


class RFBucket:
    """Holds a blueprint of the current RF bucket configuration.
    Should be requested via RFSystems.get_bucket(gamma).

    Contains all information and all physical parameters of the
    current longitudinal RF configuration for a (real, not macro-)
    particle.

    Use for plotting or obtaining the Hamiltonian etc.

    Warning: zmin and zmax do not (yet) account for phi_offset of
    Kick objects, i.e. the left and right side of the bucket are not
    accordingly moved w.r.t. the harmonic phase offset.
    """

    """Sampling points to find zero crossings."""
    sampling_points = 1000

    def __init__(self, circumference, gamma, mass_kg,
                 charge_coulomb, alpha_array, p_increment,
                 harmonic_list, voltage_list, phi_offset_list,
                 z_offset=None, *args, **kwargs):
        '''Implements only the leading order momentum compaction factor.

        Arguments:
        - mass_kg is the mass of the particle type in the beam
        - charge_coulomb is the charge of the particle type in the beam
        - z_offset determines the centre for the bucket interval
        over which the root finding (of the electric force field to
        calibrate the separatrix Hamiltonian value to zero) is done.
        z_offset is per default determined by the zero crossing located
        closest to z == 0.
        '''

        self.charge_coulomb = charge_coulomb
        self.mass_kg = mass_kg

        self._gamma = gamma
        self._beta = np.sqrt(1 - gamma**-2)
        self._p0 = np.sqrt(gamma**2 - 1) * mass_kg * c

        self.alpha0 = alpha_array[0]
        self.p_increment = p_increment

        self.circumference = circumference
        self.h = harmonic_list
        self.V = voltage_list
        self.dphi = phi_offset_list

        """Additional electric force fields to be added on top of the
        RF electric force field.
        """
        self._add_forces = []
        """Additional electric potential energies to be added on top
        of the RF electric potential energy.
        """
        self._add_potentials = []

        zmax = self.circumference / (2*np.amin(self.h))

        if z_offset is None:
            # i_fund = np.argmin(self.h) # index of fundamental RF element
            # phi_offset = self.dphi[i_fund]
            # # account for bucket size between -pi and pi.
            # # below transition there should be no relocation of the
            # # bucket interval by an offset of pi! we only need relative
            # # offset w.r.t. normal phi setting at given gamma (0 resp. pi)
            # if self.eta0 < 0:
            #     phi_offset -= np.pi
            # z_offset = -phi_offset * self.R / self.h[i_fund]
            ### the above approach does not take into account higher harmonics!
            ### Within a 2 bucket length interval we find all zero crossings
            ### of the non-accelerated total_force and identify the outermost
            ### separatrix UFPs via their minimal (convexified) potential value
            domain_to_find_bucket_centre = np.linspace(-1.999*zmax, 1.999*zmax,
                                                       self.sampling_points)
            z0 = self.zero_crossings(
                partial(self.total_force, acceleration=False),
                domain_to_find_bucket_centre)
            convex_pot0 = (
                np.array(self.total_potential(z0, acceleration=False)) *
                np.sign(self.eta0) / self.charge_coulomb)  # charge for numerical reasons
            outer_separatrix_pot0 = np.min(convex_pot0)
            outer_separatrix_z0 = z0[np.isclose(convex_pot0,
                                                outer_separatrix_pot0)]
            # outer_separatrix_z0 should contain exactly 2 entries
            z_offset = np.mean(outer_separatrix_z0)
        self.z_offset = z_offset

        """Minimum and maximum z values on either side of the
        stationary bucket to cover the maximally possible bucket area,
        defined by the fundamental harmonic.
        (This range is always larger than the outmost unstable fix
        points of the real bucket including self.p_increment .)
        """
        self.interval = (z_offset - 1.01*zmax, z_offset + 1.01*zmax)

    @property
    def gamma(self):
        return self._gamma

    @property
    def beta(self):
        return self._beta

    @property
    def p0(self):
        return self._p0

    @property
    def deltaE(self):
        return self.p_increment * self.beta * c

    @property
    def harmonic_list(self):
        return self.h
    @harmonic_list.setter
    def harmonic_list(self, value):
        self.h = value

    @property
    def voltage_list(self):
        return self.V
    @voltage_list.setter
    def voltage_list(self, value):
        self.V = value

    @property
    def phi_offset_list(self):
        return self.dphi
    @phi_offset_list.setter
    def phi_offset_list(self, value):
        self.dphi = value

    @property
    def z_ufp(self):
        '''Return the (left-most) unstable fix point on the z axis
        within self.interval .
        '''
        try:
            return self._z_ufp
        except AttributeError:
            self._z_sfp, self._z_ufp = self._get_zsfp_and_zufp()
            return self._z_ufp

    @property
    def z_sfp(self):
        '''Return the (left-most) stable fix point on the z axis.
        within self.interval .
        '''
        try:
            return self._z_sfp
        except AttributeError:
            self._z_sfp, self._z_ufp = self._get_zsfp_and_zufp()
            return self._z_sfp

    @property
    def z_ufp_separatrix(self):
        '''Return the (left-most) unstable fix point at the outermost
        separatrix of the bucket.
        (i.e. a bucket boundary defining unstable fix point)
        '''
        if self.eta0 * self.p_increment > 0:
            # separatrix ufp right of sfp
            return self.z_ufp[-1]
        else:
            # separatrix ufp left of sfp
            return self.z_ufp[0]

    @property
    def z_sfp_extr(self):
        '''Return the (left-most) absolute extremal stable fix point
        within the bucket.
        '''
        sfp_extr_index = np.argmax(self.hamiltonian(self.z_sfp, 0,
                                                    make_convex=True))
        return self.z_sfp[sfp_extr_index]

    @property
    def z_left(self):
        '''Return the left bucket boundary within self.interval .'''
        try:
            return self._z_left
        except AttributeError:
            self._z_left, self._z_right, _ = self._get_bucket_boundaries()
            return self._z_left

    @property
    def z_right(self):
        '''Return the right bucket boundary within self.interval .'''
        try:
            return self._z_right
        except AttributeError:
            self._z_left, self._z_right, _ = self._get_bucket_boundaries()
            return self._z_right

    #@property
    #@deprecated("--> Will become z_left.\n")
    #def zleft(self):
    #    '''Return the left bucket boundary within self.interval .'''
    #    try:
    #        return self._z_left
    #    except AttributeError:
    #        self._z_left, self._z_right, _ = self._get_bucket_boundaries()
    #        return self._z_left

    #@property
    #@deprecated("--> Will become z_right.\n")
    #def zright(self):
    #    '''Return the right bucket boundary within self.interval .'''
    #    try:
    #        return self._z_right
    #    except AttributeError:
    #        self._z_left, self._z_right, _ = self._get_bucket_boundaries()
    #        return self._z_right

    @property
    def R(self):
        return self.circumference/(2*np.pi)

    # should make use of eta functionality of LongitudinalMap at some point
    @property
    def eta0(self):
        return self.alpha0 - self.gamma**-2

    @property
    def beta_z(self):
        return np.abs(self.eta0 * self.R / self.Q_s)

    #@property
    #@deprecated('--> Use Q_s instead!')
    #def Qs(self):
    #    return self.Q_s

    @property
    def Q_s(self):
        """Linear synchrotron tune for small amplitudes i.e., in the
        center of the bucket. Analytical formula neglects any
        added forces / potentials via add_fields.
        """
        hV = sum([h * self.V[i] for i, h in enumerate(self.h)])
        # if hV == 0:
        #     ix = np.argmax(self.V)
        #     hV = self.h[ix] * self.V[ix]
        return np.sqrt(self.charge_coulomb*np.abs(self.eta0)*hV /
                       (2*np.pi*self.p0*self.beta*c))

    def add_fields(self, add_forces, add_potentials):
        '''Include additional (e.g. non-RF) effects to this RFBucket.
        Use this interface for adding space charge influence etc.
        to the bucket parameters and shape.

        Arguments:
        - add_forces are additional electric force fields to be added
        on top of the RF electric force field.
        add_forces is expected to be an iterable of functions of z,
        in units of Coul*Volt/metre.
        - add_potentials are additional electric potential energies
        to be added on top of the RF electric potential energy.
        add_potentials is expected to be an iterable of functions of z,
        in units of Coul*Volt.

        Bucket shape parameters z_ufp, z_sfp, z_left and z_right are
        recalculated.
        '''
        self._add_forces += add_forces
        self._add_potentials += add_potentials
        try:
            delattr(self, "_z_ufp")
            delattr(self, "_z_sfp")
        except AttributeError:
            pass
        try:
            delattr(self, "_z_left")
            delattr(self, "_z_right")
        except AttributeError:
            pass

    # FORCE FIELDS AND POTENTIALS OF MULTI-HARMONIC ACCELERATING BUCKET
    # =================================================================
    def rf_force(self, V, h, dphi, p_increment, acceleration=True):
        def f(z):
            coefficient = np.abs(self.charge_coulomb)/self.circumference
            focusing_field = reduce(lambda x, y: x+y, [
                V_i * np.sin(-h_i*z/self.R + dphi_i)
                for V_i, h_i, dphi_i in zip(V, h, dphi)])
            if not acceleration:
                accelerating_field = 0
            else:
                accelerating_field = -(
                    p_increment*self.beta*c/self.circumference)
            return coefficient * focusing_field + accelerating_field
        return f

    def total_force(self, z, ignore_add_forces=False, acceleration=True):
        '''Return the total electric force field including
        - the acceleration offset and
        - the additional electric force fields (provided via
        self.add_nonRF_influences),
        evaluated at position z in units of Coul*Volt/metre.
        '''
        f = (self.rf_force(self.V, self.h, self.dphi,
                           self.p_increment, acceleration)(z) +
             sum(f(z) for f in self._add_forces
                 if not ignore_add_forces))
        return f


    #@deprecated('--> Replace with "rf_force(acceleration=False)" ' +
    #            'as soon as possible.\n')
    #def make_singleharmonic_force(self, V, h, dphi):
    #    '''Return the electric force field of a single harmonic
    #    RF element as a function of z in units of Coul*Volt/metre.
    #    '''
    #    def force(z):
    #        return (np.abs(self.charge_coulomb) * V / self.circumference *
    #                np.sin(h * z / self.R + dphi))
    #    return force

    #@deprecated('--> Replace with "total_force(acceleration=False)" ' +
    #            'as soon as possible.\n')
    #def make_total_force(self, ignore_add_forces=False):
    #    '''Return the stationary total electric force field of
    #    superimposed RF elements (multi-harmonics) as a function of z.
    #    Parameters are taken from RF parameters of this
    #    RFBucket instance.

    #    Adds the additional electric force fields (provided via
    #    self.add_nonRF_influences) on top.
    #    Uses units of Coul*Volt/metre.
    #    '''
    #    def total_force(z):
    #        '''Return stationary total electric force field of
    #        superimposed RF elements (multi-harmonics) and additional
    #        force fields as a function of z in units of Coul*Volt/metre.
    #        '''
    #        harmonics = (self.make_singleharmonic_force(V, h, dphi)(z)
    #                     for V, h, dphi in zip(self.V, self.h, self.dphi))
    #        return (sum(harmonics) + sum(f(z) for f in self._add_forces
    #                                     if not ignore_add_forces))
    #    return total_force

    #@deprecated('--> Replace with "total_force" as soon as possible.\n')
    #def acc_force(self, z, ignore_add_forces=False):
    #    '''Return the total electric force field including
    #    - the acceleration offset and
    #    - the additional electric force fields (provided via
    #    self.add_nonRF_influences),
    #    evaluated at position z in units of Coul*Volt/metre.
    #    '''
    #    total_force = self.make_total_force(
    #        ignore_add_forces=ignore_add_forces)
    #    return total_force(z) - self.deltaE / self.circumference


    def rf_potential(self, V, h, dphi, p_increment,
                     acceleration=True, offset=True):
        '''Return the RF electric potential energy including the linear
        acceleration slope (if acceleration == True).

        Arguments:
            - V: list of voltages for each harmonic
            - h: list of harmonics
            - dphi: list of phase offsets for each harmonic
            - p_increment: momentum increase per turn
            - acceleration: whether to superimpose the linear
              acceleration slope (induced by p_increment, default=True)
            - offset: boolean whether the potential energy should be
              shifted to zero at the unstable fix point enclosing
              the separatrix of the RF bucket (default=True).
        '''
        def vf(z):
            coefficient = np.abs(self.charge_coulomb)/self.circumference
            focusing_potential = reduce(lambda x, y: x+y, [
                -self.R/h[i] * V[i] * np.cos(-h[i]*z/self.R + dphi[i])
                for i in range(len(V))])
            return coefficient * focusing_potential

        if not acceleration:
            return vf
        else:
            v_norm = 0 # normalisation shift
            if offset:
                zmax = self.z_ufp_separatrix
                v_norm = (vf(zmax) +
                          p_increment*self.beta*c/self.circumference * zmax)

            def f(z):
                return (vf(z) + p_increment*self.beta*c/self.circumference * z
                        - v_norm)
            return f

    def total_potential(self, z, ignore_add_potentials=False,
                        make_convex=False, acceleration=True, offset=True):
        '''Return the total electric potential energy including
        - the linear acceleration slope and
        - the additional electric potential energies (provided via
        self.add_nonRF_influences),
        evaluated at position z in units of Coul*Volt.

        Note:
        Adds a potential energy offset: this relocates the extremum
        (defining the unstable fix point UFP of the bucket)
        to obtain zero potential energy at the UFP.
        Thus the Hamiltonian value of the separatrix is calibrated
        to zero.

        Arguments:
            - make_convex: multiplies by sign(eta) for plotting etc.
              To see a literal 'bucket structure' in the sense of
              always having a local maximum in the Hamiltonian topology
              where the stable fix points are located, set
              make_convex=True in order to return
              sign(eta)*hamiltonian(z, dp).
            - offset: boolean whether the potential energy should be
              shifted to zero at the unstable fix point enclosing
              the separatrix of the RF bucket (default=True).
        '''
        v = (self.rf_potential(self.V, self.h, self.dphi,
                               self.p_increment, acceleration, offset)(z) +
             sum(pot(z) for pot in self._add_potentials
                 if not ignore_add_potentials))
        # if make_convex:
        #     v *= np.sign(self.eta0)
        return v

    # ROOT AND BOUNDARY FINDING ROUTINES
    # ==================================
    def zero_crossings(self, f, x=None, subintervals=None):
        '''Determine roots of f along x.
        If x is not explicitely given, take stationary bucket interval.
        '''
        if x is None:
            if subintervals is None:
                subintervals = self.sampling_points
            x = np.linspace(*self.interval, num=subintervals)

        return cvt_zero_crossings(f, x)

    def _get_bucket_boundaries(self):
        '''Return the bucket boundaries as well as the whole list
        of acceleration voltage roots, (z_left, z_right, z_roots).
        '''
        z0 = np.atleast_1d(self.zero_crossings(self.total_potential))
        z0 = np.append(z0, self.z_ufp)
        return np.min(z0), np.max(z0), z0

    def _get_zsfp_and_zufp(self):
        '''Return (z_sfp, z_ufp),
        where z_sfp is the z location of the stable fix points,
        and z_ufp is the z location of the unstable fix points
        belonging to this RF bucket.

        A stationary RF bucket has the right-most UFP overlapping
        with the adjacent RF bucket's separatrix, while for an
        ac-/decelerating RF bucket one of the two out-most UFP always
        belongs to the separatrix of the adjacent RF bucket. This UFP
        will be discarded (although you find it with the zero crossing
        of the total_force) by comparing the voltages between the
        out-most UFP.
        '''
        z0 = np.atleast_1d(self.zero_crossings(self.total_force))

        if not z0.size:
            # no bucket (i.e. bucket area 'negative')
            raise ValueError('With an electric force field this weak ' +
                             'there is no bucket for such strong ' +
                             'momentum increase -- ' +
                             'why do you ask me for bucket boundaries ' +
                             'in this hyperbolic phase space structure?!')

        if len(z0) == 1:  # exactly zero bucket area
            return z0, z0

        V_left = self.total_potential(z0[0], make_convex=True, offset=False)
        V_right = self.total_potential(z0[-1], make_convex=True, offset=False)

        if self.eta0 * self.p_increment > 0:
            # separatrix ufp right of sfp AND we are ac-/decelerating
            if V_left < V_right:
                # --> need to remove first ufp (belongs to bucket to the left)
                z0 = z0[1:]
            z_sfp, z_ufp = z0[::2], z0[1::2]
        elif self.eta0 * self.p_increment == 0:
            # stationary bucket, need both left and right UFP (overlapping!)
            z_sfp, z_ufp = z0[1::2], z0[::2]
        else:
            # separatrix ufp left of sfp AND we are ac-/decelerating
            if V_right < V_left:
                # --> need to remove last ufp (belongs to bucket to the right)
                z0 = z0[:-1]
            z_sfp, z_ufp = z0[1::2], z0[::2]

        return z_sfp, z_ufp

    # HAMILTONIANS, SEPARATRICES AND RELATED FUNCTIONS
    # ================================================
    def hamiltonian(self, z, dp, make_convex=False):
        '''Return the Hamiltonian at position z and dp in units of
        Coul*Volt/p0.

        Arguments:
            - make_convex: multiplies by sign(eta) for plotting etc.
              To see a literal 'bucket structure' in the sense of
              always having a local maximum in the Hamiltonian topology
              where the stable fix points are located, set
              make_convex=True in order to return
              sign(eta)*hamiltonian(z, dp).
        '''
        h = (-0.5 * self.eta0 * self.beta * c * dp**2 +
             self.total_potential(z) / self.p0)
        # if make_convex:
        #     h *= np.sign(self.eta0)
        return h

    def equihamiltonian(self, zcut):
        '''Return a function dp_at that encodes the equi-Hamiltonian
        contour line that cuts the z axis at (zcut, 0).
        In more detail, dp_at(z) returns the (positive) dp value at
        its given z argument such that
        self.hamiltonian(z, dp_at(z)) == self.hamiltonian(zcut, 0) .
        '''
        def dp_at(z):
            hcut = self.hamiltonian(zcut, 0)
            r = np.abs(2./(self.eta0*self.beta*c) *
                       (self.total_potential(z)/self.p0 - hcut))
            return np.sqrt(r.clip(min=0))
        return dp_at

    def separatrix(self, z):
        '''Return the positive dp value corresponding to the separatrix
        Hamiltonian contour line at the given z.
        '''
        dp_separatrix_at = self.equihamiltonian(self.z_ufp_separatrix)
        return dp_separatrix_at(z)

    def h_sfp(self, make_convex=False):
        '''Return the extremal Hamiltonian value at the corresponding
        stable fix point (self.z_sfp_extr, 0) of the bucket.
        '''
        return self.hamiltonian(self.z_sfp_extr, 0, make_convex)

    def dp_max(self, zcut):
        '''Return the maximal dp value along the equihamiltonian which
        is located at (one of the) self.z_sfp .
        '''
        dp_at = self.equihamiltonian(zcut)
        return np.amax(dp_at(self.z_sfp))

    def is_in_separatrix(self, z, dp, margin=0):
        """Return boolean whether the coordinate (z, dp) is located
        strictly inside the separatrix of this bucket
        (i.e. excluding neighbouring buckets).

        If margin is different from 0, use the equihamiltonian
        defined by margin*self.h_sfp instead of the separatrix.
        (Use margin as a weighting factor in units of the Hamiltonian
        value at the stable fix point to move from the separatrix
        toward the extremal Hamiltonian value at self.z_sfp .)
        """
        within_interval = np.logical_and(self.z_left < z, z < self.z_right)
        within_separatrix = (self.hamiltonian(z, dp, make_convex=True) >
                             margin * self.h_sfp(make_convex=True))
        return np.logical_and(within_interval, within_separatrix)

    def make_is_accepted(self, margin=0):
        """Return the function is_accepted(z, dp) definining the
        equihamiltonian with a value of margin*self.h_sfp .
        For margin 0, the returned is_accepted(z, dp) function is
        identical to self.is_in_separatrix(z, dp).
        """
        return partial(self.is_in_separatrix, margin=margin)

    def emittance_single_particle(self, z=None, sigma=2):
        """The single particle emittance computed along a given
        equihamiltonian line.
        """
        if z is not None:
            zl = -sigma * z
            zr = +sigma * z
            f = self.equihamiltonian(sigma * z)
        else:
            zl = self.z_left
            zr = self.z_right
            f = self.separatrix

        Q, error = dblquad(lambda y, x: 1, zl, zr,
                           lambda x: 0, f)

        return Q * 2*self.p0/np.abs(self.charge_coulomb)

    def bunchlength_single_particle(self, epsn_z, verbose=False):
        """The corresponding RMS bunch length computed from the single
        particle emittance.
        """
        def emittance_from_zcut(zcut):
            emittance = self.emittance_single_particle(zcut)
            if np.isnan(emittance):
                raise ValueError

            return emittance - epsn_z

        sigma = newton(emittance_from_zcut, 1)

        return sigma

    def guess_H0(self, var, from_variable='epsn', make_convex=True):
        """Pure estimate value of H_0 starting from a bi-Gaussian bunch
        in a linear "RF bucket". Intended for use by iterative matching
        algorithms in the generators module.
        """
        # If Qs = 0, get the fundamental harmonic
        hV = sum([h * V for h, V in zip(self.h, self.V)])
        if hV == 0:
            ix = np.argmax(self.V)
            hV = self.h[ix] * self.V[ix]
        Qs = np.sqrt(np.abs(self.charge_coulomb)*np.abs(self.eta0)*hV /
                     (2*np.pi*self.p0*self.beta*c))
        beta_z = np.abs(self.eta0 * self.R / Qs)

        # to be replaced with something more flexible (add_forces etc.)
        if from_variable == 'epsn':
            epsn = var
            z0 = np.sqrt(epsn/(4.*np.pi) * beta_z *
                         np.abs(self.charge_coulomb)/self.p0)  # gauss approx.
        elif from_variable == 'sigma':
            z0 = var

        h0 = self.beta*c * (z0/beta_z)**2
        # if make_convex:
        h0 *= np.abs(self.eta0)
        return h0



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import numpy as np
import pytest
import xtrack as xt
import xobjects as xo

import xpart as xp
from xobjects.test_helpers import fix_random_seed


@pytest.mark.parametrize('scenario', ['psb_injection', 'sps_ions', 'lhc_protons'])
@pytest.mark.parametrize('distribution', ['gaussian', 'parabolic'])
@fix_random_seed(46398498)
def test_single_rf_harmonic_matcher_rms_and_profile_and_tune(
                                    scenario, distribution):
    for ctx in xo.context.get_test_contexts():
        print(f"Test {ctx.__class__}")

        if scenario == "lhc_protons":
            # Build a reference particle
            p0 = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1, p0c=7e12, x=1, y=3,
                              delta=[10], _context=ctx)

            # Load machine model (from pymask)
            filename = xt._pkg_root.parent.joinpath(
                'test_data/lhc_no_bb/line_and_particle.json')
            with open(filename, 'r') as fid:
                input_data = json.load(fid)
            line = xt.Line.from_dict(input_data['line'])
            line.build_tracker(_context=ctx)
            line.particle_ref = p0
            rms_bunch_length = 0.10

        elif scenario == "sps_ions":
            # Load machine model (spsion)
            filename = xt._pkg_root.parent.joinpath(
                'test_data/sps_ions/line_and_particle.json')
            with open(filename, 'r') as fid:
                input_data = json.load(fid)
            line = xt.Line.from_dict(input_data)
            line.build_tracker(_context=ctx)
            rms_bunch_length = 0.10
        elif scenario == "psb_injection":
            # Load machine model (psb injection)
            filename = xt._pkg_root.parent.joinpath(
                'test_data/psb_injection/line_and_particle.json')
            with open(filename, 'r') as fid:
                input_data = json.load(fid)
            line = xt.Line.from_dict(input_data)
            line.build_tracker(_context=ctx)
            rms_bunch_length = 17.
        else:
            raise NotImplementedError

        zeta, delta, matcher = xp.generate_longitudinal_coordinates(
            line=line, # particle_ref=p0,
            num_particles=1000000,
            sigma_z=rms_bunch_length, distribution=distribution,
            engine="single-rf-harmonic", return_matcher=True)
        tau = zeta / line.particle_ref._xobject.beta0[0]
        tau_distr_y = matcher.tau_distr_y
        tau_distr_x = matcher.tau_distr_x
        dx = tau_distr_x[1] - tau_distr_x[0]
        hist, _ = np.histogram(tau,
                    range=(tau_distr_x[0]-dx/2., tau_distr_x[-1]+dx/2.),
                    bins=len(tau_distr_x))
        hist = hist / sum(hist) * sum(tau_distr_y)

        twiss_tune = line.twiss()['qs']
        theoretical_synchrotron_tune = matcher.get_synchrotron_tune()

        assert np.isclose(theoretical_synchrotron_tune,
                          twiss_tune, rtol=3.e-3, atol=1.e-15)

        assert np.isclose(rms_bunch_length, np.std(zeta),
                          rtol=2e-2, atol=1e-15)

        xo.assert_allclose(hist, tau_distr_y, atol=3.e-2, rtol=1.e-2)



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from warnings import warn

# This file is deprecated and will be removed in the future.


def __getattr__(name):
    if name.startswith('__'):
        # Needed for module resolution, i.e. for the below errors to actually
        # work properly
        raise AttributeError

    warn("The functionality in xpart.pdg moved to xtrack.particles.pdg. Using it "
      + f"via xpart is still possible, but deprecated. It will be removed in the "
      + f"future.", FutureWarning)
    import xtrack.particles.pdg as pdg
    if name in pdg.__dict__:
        return pdg.__dict__[name]
    else:
        raise AttributeError(f"module 'xtrack.particles.pdg' has no " \
                           + f"attribute '{name}'")




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

def __getattr__(name):
    if name.startswith('__'):
        # Needed for module resolution, i.e. for the below errors to actually
        # work properly
        raise AttributeError

    if name == 'ParticlesPurelyLongitudinal':
        raise ModuleNotFoundError(
            "ParticlesPurelyLongitudinal class has been removed. Please track "
            "the usual Particles class with the config flags `FREEZE_VAR_x`, "
            "etc. set on the line instead."
        )

    raise ModuleNotFoundError(
        "It seems you are trying to import a Particles related object from "
        "xpart.particles. Please note that the Particles class has been moved "
        "to Xtrack, and can be imported as xtrack.Particles."
    )



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from .generate_longitudinal import (generate_longitudinal_coordinates,
                                    _characterize_line, get_bucket)
from .single_rf_harmonic_matcher import SingleRFHarmonicMatcher
from .generate_binomial_longitudinal_distribution import generate_binomial_longitudinal_coordinates
from .generate_parabolic_longitudinal_distribution import generate_parabolic_longitudinal_coordinates
from .generate_qgaussian_longitudinal_distribution import generate_qgaussian_longitudinal_coordinates



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

from .general import _print

from .longitudinal import generate_longitudinal_coordinates, _characterize_line
from .build_particles import build_particles

# To get the right Particles class depending on pyheatail interface state
import xpart as xp

def generate_matched_gaussian_bunch(num_particles,
                                    nemitt_x, nemitt_y, sigma_z,
                                    total_intensity_particles=None,
                                    particle_on_co=None,
                                    R_matrix=None,
                                    circumference=None,
                                    momentum_compaction_factor=None,
                                    rf_harmonic=None,
                                    rf_voltage=None,
                                    rf_phase=None,
                                    energy_ref_increment=None,
                                    tracker=None,
                                    line=None,
                                    particle_ref=None,
                                    engine=None,
                                    return_matcher=False,
                                    _context=None, _buffer=None, _offset=None,
                                    **kwargs,  # Passed to build_particles
                                    ):
    '''
    Generate a matched Gaussian bunch.

    Parameters
    ----------
    line : xpart.Line
        Line for which the bunch is generated.
    num_particles : int
        Number of particles to be generated.
    nemitt_x : float
        Normalized emittance in the horizontal plane (in m rad).
    nemitt_y : float
        Normalized emittance in the vertical plane (in m rad).
    sigma_z : float
        RMS bunch length in meters.
    total_intensity_particles : float
        Total intensity of the bunch in particles.

    Returns
    -------
    part : xpart.Particles
        Particles object containing the generated particles.

    '''

    if line is not None and tracker is not None:
        raise ValueError(
            'line and tracker cannot be provided at the same time.')

    if tracker is not None:
        _print(
            "The argument tracker is deprecated. Please use line instead.",
            DeprecationWarning)
        line = tracker.line

    if line is not None and line.tracker is None:
        line.build_tracker()

    if particle_ref is not None and particle_on_co is not None:
        raise ValueError("`particle_ref` and `particle_on_co`"
                         " cannot be provided at the same time")

    if particle_ref is None:
        if particle_on_co is not None:
            particle_ref = particle_on_co
        elif line is not None and line.particle_ref is not None:
            particle_ref = line.particle_ref
        else:
            raise ValueError(
                "`line`, `particle_ref` or `particle_on_co` must be provided!")

    zeta, delta, matcher = generate_longitudinal_coordinates(
        distribution='gaussian',
        num_particles=num_particles,
        particle_ref=(particle_ref if particle_ref is not None
                      else particle_on_co),
        line=line,
        circumference=circumference,
        momentum_compaction_factor=momentum_compaction_factor,
        rf_harmonic=rf_harmonic,
        rf_voltage=rf_voltage,
        rf_phase=rf_phase,
        energy_ref_increment=energy_ref_increment,
        sigma_z=sigma_z,
        engine=engine,
        return_matcher=True,
        **kwargs)

    assert len(zeta) == len(delta) == num_particles

    x_norm = np.random.normal(size=num_particles)
    px_norm = np.random.normal(size=num_particles)
    y_norm = np.random.normal(size=num_particles)
    py_norm = np.random.normal(size=num_particles)

    if total_intensity_particles is None:
        # go to particles.weight = 1
        total_intensity_particles = num_particles

    part = build_particles(_context=_context, _buffer=_buffer, _offset=_offset,
                      R_matrix=R_matrix,
                      particle_on_co=particle_on_co,
                      particle_ref=(
                          particle_ref if particle_on_co is  None else None),
                      line=line,
                      zeta=zeta, delta=delta,
                      x_norm=x_norm, px_norm=px_norm,
                      y_norm=y_norm, py_norm=py_norm,
                      nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                      weight=total_intensity_particles/num_particles,
                      **kwargs)
    if return_matcher:
        return part, matcher
    else:
        return part


def split_scheme(filling_scheme, n_chunk=1):
    """
    Distribute the filling scheme between the processes, i.e. assign to each
    processor its bunches
    """
    total_n_bunches = len(filling_scheme.nonzero()[0])
    if n_chunk > 1:

        # create the array containing the id of the bunches on each rank
        # (copied from PyHEADTAIL.mpi.mpi_data)
        n_bunches = total_n_bunches
        n_bunches_on_rank = [n_bunches//n_chunk+1 if i < n_bunches % n_chunk
                             else n_bunches // n_chunk + 0
                             for i in range(n_chunk)]
        n_tasks_cumsum = np.insert(np.cumsum(n_bunches_on_rank), 0, 0)
        total_bunch_ids = np.unique(
            np.cumsum(filling_scheme == 1)) - 1
        bunches_per_rank = [total_bunch_ids[n_tasks_cumsum[i]:
                                            n_tasks_cumsum[i + 1]]
                            for i in range(n_chunk)]
    else:
        bunches_per_rank = [np.linspace(0,
                                        total_n_bunches - 1,
                                        total_n_bunches)]

    bunches_per_rank = list(map(np.int64, bunches_per_rank))

    return bunches_per_rank


def generate_matched_gaussian_multibunch_beam(filling_scheme,
                                              bunch_num_particles,
                                              nemitt_x, nemitt_y, sigma_z,
                                              bunch_intensity_particles=None,
                                              particle_on_co=None,
                                              R_matrix=None,
                                              circumference=None,
                                              momentum_compaction_factor=None,
                                              rf_harmonic=None,
                                              rf_voltage=None,
                                              rf_phase=None,
                                              bucket_length = None,
                                              energy_ref_increment=None,
                                              tracker=None,
                                              line=None,
                                              particle_ref=None,
                                              engine=None,
                                              _context=None, _buffer=None, _offset=None,
                                              bunch_selection=None,
                                              bunch_spacing_buckets=1,
                                              prepare_line_and_particles_for_mpi_wake_sim=False,
                                              communicator=None,
                                              **kwargs,  # Passed to build_particles
                                              ):

    if particle_ref is None and line is not None:
        particle_ref = line.particle_ref

    assert ((line is not None and particle_ref is not None) or
            (rf_harmonic is not None and rf_voltage is not None) or
            bucket_length is not None)

    if circumference is None:
        circumference = line.get_length()
    assert circumference > 0.0

    if bucket_length is not None:
        assert (rf_harmonic is None and rf_voltage is None),(
                'Cannot provide bucket length together with RF voltage+harmonic')
    else:
        if rf_harmonic is not None and rf_voltage is not None:
            main_harmonic_number = rf_harmonic[np.argmax(rf_voltage)]
        else:
            dct_line = _characterize_line(line, particle_ref)
            assert len(dct_line['voltage_list']) > 0
            main_harmonic_number = int(np.floor(
                        dct_line['h_list'][np.argmax(dct_line['voltage_list'])]+0.5))
        bucket_length = circumference/main_harmonic_number
    bunch_spacing = bunch_spacing_buckets * bucket_length
    assert filling_scheme is not None
    assert len(filling_scheme) <= np.floor(circumference/bunch_spacing+0.5)

    if len(filling_scheme) < np.floor(circumference/bunch_spacing+0.5):
        filling_scheme = np.concatenate(
            (filling_scheme,
            np.zeros(int(np.floor(circumference/bunch_spacing+0.5) - len(filling_scheme)),
                    dtype=np.int64)))

    if prepare_line_and_particles_for_mpi_wake_sim and bunch_selection is None:
        if communicator is None:
            from mpi4py import MPI
            communicator = MPI.COMM_WORLD

        if communicator.Get_size() <= 1:
            raise ValueError('when `prepare_line_and_particles_for_mpi_wake_sim` is True, '
                             'MPI communicator must have more than one rank')
        bunch_selection_rank = split_scheme(filling_scheme=filling_scheme,
                                             n_chunk=int(communicator.Get_size()))
        bunch_selection = bunch_selection_rank[communicator.Get_rank()]

    if bunch_selection is None:
        bunch_selection = range(len(filling_scheme.nonzero()[0]))

    macro_bunch = generate_matched_gaussian_bunch(
        num_particles=bunch_num_particles * len(bunch_selection),
        nemitt_x=nemitt_x, nemitt_y=nemitt_y, sigma_z=sigma_z,
        total_intensity_particles=bunch_intensity_particles * len(bunch_selection),
        particle_on_co=particle_on_co,
        R_matrix=R_matrix,
        circumference=circumference,
        momentum_compaction_factor=momentum_compaction_factor,
        rf_harmonic=rf_harmonic,
        rf_voltage=rf_voltage,
        rf_phase=rf_phase,
        energy_ref_increment=energy_ref_increment,
        tracker=tracker,
        line=line,
        particle_ref=particle_ref,
        engine=engine,
        _context=_context, _buffer=_buffer, _offset=_offset,
        **kwargs,  # They are passed to build_particles
    )

    filled_buckets = filling_scheme.nonzero()[0]
    count = 0
    for bunch_number in bunch_selection:
        bucket_n = filled_buckets[bunch_number]
        macro_bunch.zeta[count * bunch_num_particles:
                         (count+1) * bunch_num_particles] -= (bunch_spacing *
                                                        bucket_n)
        count += 1

    if prepare_line_and_particles_for_mpi_wake_sim:
        import xwakes as xw
        xw.config_pipeline_for_wakes(
            particles=macro_bunch,
            line=line,
            communicator=communicator)

    return macro_bunch



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

from xtrack.particles import (Particles, PROTON_MASS_EV, ELECTRON_MASS_EV,
                              MUON_MASS_EV, Pb208_MASS_EV, reference_from_pdg_id,
                              enable_pyheadtail_interface, disable_pyheadtail_interface)
pmass = PROTON_MASS_EV  # backwards compatibility

from xtrack.particles.pdg import get_pdg_id_from_name, get_name_from_pdg_id

from .build_particles import build_particles
from .matched_gaussian import (generate_matched_gaussian_bunch,
                               generate_matched_gaussian_multibunch_beam,
                               )

from .transverse_generators import generate_2D_polar_grid
from .transverse_generators import generate_2D_uniform_circular_sector
from .transverse_generators import generate_2D_pencil
from .transverse_generators import generate_2D_pencil_with_absolute_cut
from .transverse_generators import generate_2D_gaussian
from .transverse_generators import (generate_hypersphere_2D, generate_hypersphere_4D,
                                    generate_hypersphere_6D)

from .longitudinal import generate_longitudinal_coordinates
from .longitudinal.generate_longitudinal import _characterize_line

from .monitors import PhaseMonitor

from ._version import __version__



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

class PhaseMonitor:
    def __init__(self, tracker=None, num_particles=1, twiss=None, line=None):

        if line is not None:
            assert tracker is None
            assert line.tracker is not None
            tracker = line.tracker

        self.twiss = twiss
        self.phase_x = []
        self.phase_y = []
        import xtrack as xt # Avoid circular dependency between xtrack as xpart
        self._monitor = xt.ParticlesMonitor(_context=tracker._buffer.context,
                                         num_particles=num_particles,
                                         start_at_turn=0,
                                         stop_at_turn=1)


    def measure(self, particles):
        self._monitor.track(particles)

        delta = self._monitor.delta
        tw = self.twiss

        for ss in 'xy':
            r = getattr(self._monitor, ss).copy()
            pr = getattr(self._monitor, f'p{ss}').copy()

            r -= delta*tw[f'd{ss}'][0]
            pr -= delta*tw[f'dp{ss}'][0]

            betr = tw[f'bet{ss}'][0]
            alfr = tw[f'alf{ss}'][0]

            phase = np.angle(r / np.sqrt(betr) -
                       1j*(r * alfr / np.sqrt(betr) +
                       pr * np.sqrt(betr)))

            getattr(self, f'phase_{ss}').append(
                              np.atleast_1d(np.squeeze(phase)))

        self._monitor.start_at_turn += 1
        self._monitor.stop_at_turn += 1

    @property
    def qx(self):
        return np.mod(np.diff(np.array(self.phase_x), axis=0)/(2*np.pi), 1)

    @property
    def qy(self):
        return np.mod(np.diff(np.array(self.phase_y), axis=0)/(2*np.pi), 1)



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #

import numpy as np
import xpart as xp

from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_reorganize_particles(test_context):
    c2n = test_context.nparray_from_context_array
    n2c = test_context.nparray_to_context_array

    particles = xp.Particles(_context=test_context, p0c=450e9, x=[1, 2, 3, 4, 5],
                             delta=[0.1, 0.2, 0.3, 0.4, 0.5])
    particles0 = particles.copy()

    assert np.all(c2n(particles.state) == 1)

    n_act, n_lost = particles.reorganize()
    assert n_act == 5

    particles.state = n2c(np.array([1, 1, 0, 0, 1]))
    n_act, n_lost = particles.reorganize()
    assert n_act == 3
    assert n_lost == 2

    assert np.all(c2n(particles.x[0:3]) == [1, 2, 5])
    assert np.all(c2n(particles.delta[0:3]) == [0.1, 0.2, 0.5])
    assert np.all(c2n(particles.ptau[0:3]) == c2n(particles0.ptau)[np.array([0, 1, 4])])


@for_all_test_contexts
def test_hide_lost_particles(test_context):
    c2n = test_context.nparray_from_context_array
    n2c = test_context.nparray_to_context_array

    particles = xp.Particles(_context=test_context, p0c=450e9, x=[1, 2, 3, 4, 5],
                             delta=[0.1, 0.2, 0.3, 0.4, 0.5])
    particles0 = particles.copy()

    assert np.all(c2n(particles.state) == 1)

    particles.hide_lost_particles()
    assert(len(particles.x) == 5)
    particles.unhide_lost_particles()
    assert(len(particles.x) == 5)

    particles.state = n2c(np.array([1, 1, 0, 0, 1]))
    particles.hide_lost_particles()
    assert(len(particles.x) == 3)
    assert np.all(c2n(particles.x) == [1, 2, 5])
    assert np.all(c2n(particles.delta) == [0.1, 0.2, 0.5])
    assert np.all(c2n(particles.ptau) == c2n(particles0.ptau)[np.array([0, 1, 4])])

    particles.unhide_lost_particles()
    assert(len(particles.x) == 5)
    assert np.all(c2n(particles.x[0:3]) == [1, 2, 5])
    assert np.all(c2n(particles.delta[0:3]) == [0.1, 0.2, 0.5])
    assert np.all(c2n(particles.ptau[0:3]) == c2n(particles0.ptau)[np.array([0, 1, 4])])



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np
from scipy import stats

import xpart as xp


def test_hypersphere_2D():
    num_particles = int(20e4)
    r = 2.5

    means = []
    stds = []
    for seed in [1, 2, 3]:
        x_norm, px_norm = xp.generate_hypersphere_2D(
            num_particles, r=r, rng_seed=seed)

        data = np.array([x_norm, px_norm]).T
        bins = [np.linspace(-r, r, 50) for i in range(2)]

        bincounts = stats.binned_statistic_dd(
            data, 1, bins=bins, statistic='count').statistic
        bincounts = bincounts[bincounts != 0]

        means.append(np.mean(bincounts))
        stds.append(np.std(bincounts))

    assert np.allclose(np.diff(means)/np.mean(means), 0, rtol=0, atol=1e-2)
    assert np.allclose(np.array(stds)/np.array(means), 0, rtol=0, atol=0.5)


def test_hypersphere_4D():
    num_particles = int(20e4)
    rx = 2.5
    ry = 4.2

    means = []
    stds = []
    for seed in [1, 2, 3]:
        x_norm, px_norm, y_norm, py_norm = xp.generate_hypersphere_4D(
            num_particles, rx=rx, ry=ry, rng_seed=seed)

        data = np.array([x_norm, px_norm, y_norm, py_norm]).T
        bins = [np.linspace(-rx, rx, 50) for i in range(2)] + \
            [np.linspace(-ry, ry, 50) for i in range(2)]

        bincounts = stats.binned_statistic_dd(
            data, 1, bins=bins, statistic='count').statistic
        bincounts = bincounts[bincounts != 0]

        means.append(np.mean(bincounts))
        stds.append(np.std(bincounts))

    assert np.allclose(np.diff(means)/np.mean(means), 0, rtol=0, atol=1e-2)
    assert np.allclose(np.array(stds)/np.array(means), 0, rtol=0, atol=0.5)


def test_hypersphere_6D():
    num_particles = int(20e4)
    rx = 2.5
    ry = 4.2
    rzeta = 1.2

    means = []
    stds = []
    for seed in [1, 2, 3]:
        x_norm, px_norm, y_norm, py_norm, zeta_norm, pzeta_norm = xp.generate_hypersphere_6D(
            num_particles, rx=rx, ry=ry, rzeta=rzeta, rng_seed=seed)

        data = np.array(
            [x_norm, px_norm, y_norm, py_norm, zeta_norm, pzeta_norm]).T
        bins = ([np.linspace(-rx, rx, 20) for i in range(2)]
                + [np.linspace(-ry, ry, 20) for i in range(2)]
                + [np.linspace(-rzeta, rzeta, 20) for i in range(2)])

        bincounts = stats.binned_statistic_dd(
            data, 1, bins=bins, statistic='count').statistic
        bincounts = bincounts[bincounts != 0]

        means.append(np.mean(bincounts))
        stds.append(np.std(bincounts))

    assert np.allclose(np.diff(means)/np.mean(means), 0, rtol=0, atol=1e-2)
    assert np.allclose(np.array(stds)/np.array(means), 0, rtol=0, atol=0.5)



from xpart.longitudinal import generate_longitudinal_coordinates
import numpy as np
from ..general import _print

def generate_parabolic_longitudinal_coordinates(num_particles=None,
									    		nemitt_x=None, 
												nemitt_y=None, 
												sigma_z=None,
												particle_ref=None, 
												tracker=None,
												line=None,
												return_matcher=False
												):

	"""
	Function to generate a parabolic longitudinal distribution 

	Parameters:
	-----------
	num_particles : int
		number of macroparticles
	nemitt_x : float
		normalized horizontal emittance in m rad
    nemitt_y : float
		normalized vertical emittance in m rad
	sigma_z : float
		bunch length in meters
	particle_ref : xp.particle
		reference particle
	tracker : xt.tracker
	line: xt.line
	return_matcher : bool
		whether to also return xp.SingleRFHarmonicMatcher object

	Returns:
	-------- 
	zeta : np.ndarray
		longitudinal coordinates zeta for particles
	delta : np.ndarray
		relative momentum offset coordinates for particles
	matcher : xp.SingleRFHarmonicMatcher
		RF matcher object
	"""
	
	if line is not None and tracker is not None:
		raise ValueError(
		    'line and tracker cannot be provided at the same time.')

	if tracker is not None:
		_print('Warning! '
		    "The argument tracker is deprecated. Please use line instead.")
		line = tracker.line

	if line is not None:
		assert line.tracker is not None, ("The line must have a tracker, "
		    "please call Line.build_tracker() first.")
	
	if num_particles is None:
		raise ValueError(
				'Number of particles must be provided')
	if sigma_z is None:
		raise ValueError(
				'Bunch length sigma_z must be provided')
				
	if particle_ref is None:
		raise ValueError(
				'Reference particle must be provided')
	
	# If emittances are not provided, set them to default value of one
	if nemitt_x is None:
		nemitt_x = 1.0
	
	if nemitt_y is None:
		nemitt_y = 1.0
	
	# Generate longitudinal coordinates s
	zeta, delta, matcher = generate_longitudinal_coordinates(line=line, distribution='parabolic', 
							num_particles=num_particles, 
							engine='single-rf-harmonic', sigma_z=sigma_z,
							particle_ref=particle_ref, return_matcher=True)
	
	if return_matcher:
		return zeta, delta, matcher
	else:
		return zeta, delta



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xpart as xp


def test_generate_2D_polar_grid():
    x_norm, px_norm, r_points, theta_points = xp.generate_2D_polar_grid(
            r_range=(0.1, 10), nr=8, theta_range=(0, np.pi/2), ntheta=10)

    assert len(x_norm) == len(px_norm) == len(r_points) == len(theta_points) == 80

    assert np.isclose(np.max(x_norm), 10)
    assert np.isclose(np.max(px_norm), 10)
    assert np.isclose(np.min(np.sqrt(x_norm**2 + px_norm**2)), 0.1)

    assert np.isclose(np.min(theta_points), 0)
    assert np.isclose(np.max(theta_points), np.pi/2)
    assert np.isclose(np.min(r_points), 0.1)
    assert np.isclose(np.max(r_points), 10)




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import logging

import numpy as np

import xobjects as xo
from .general import _print

import xtrack as xt

logger = logging.getLogger(__name__)

def _check_lengths(**kwargs):
    length = None
    for nn, xx in kwargs.items():
        if hasattr(xx, "__iter__"):
            if hasattr(xx, 'shape') and len(xx.shape) == 0:
                continue
            if length is None:
                length = len(xx)
            else:
                if length != len(xx):
                    raise ValueError(f"invalid length len({nn})={len(xx)}")
    if 'num_particles' in kwargs.keys():
        num_particles = kwargs['num_particles']
        if num_particles is not None and length is None:
            length = num_particles
        if num_particles is not None and length != num_particles:
            raise ValueError(
              f"num_particles={num_particles} is inconsistent with array length")

    if length is None:
        length = 1
    return length

def build_particles(_context=None, _buffer=None, _offset=None, _capacity=None,
                      mode=None,
                      particle_ref=None,
                      num_particles=None,
                      x=None, px=None, y=None, py=None,
                      zeta=None, delta=None, pzeta=None, ptau=None,
                      x_norm=None, px_norm=None, y_norm=None, py_norm=None,
                      zeta_norm=None, pzeta_norm=None,
                      tracker=None,
                      line=None,
                      at_element=None,
                      match_at_s=None,
                      particle_on_co=None,
                      R_matrix=None,
                      W_matrix=None,
                      method=None,
                      nemitt_x=None, nemitt_y=None,nemitt_zeta = None,
                      scale_with_transverse_norm_emitt=None,
                      weight=None,
                      s_tol=1e-6,
                      include_collective=False,
                      **kwargs, # They are passed to the twiss
                    ):

    """
    Same as `xtrack.Line.build_particles`. See there for documentation.

    """

    if line is not None and tracker is not None:
        raise ValueError(
            'line and tracker cannot be provided at the same time.')

    if tracker is not None:
        _print('Warning! '
            "The argument tracker is deprecated. Please use line instead.")
        line = tracker.line

    if line is not None:
        assert line.tracker is not None, ("The line must have a tracker, "
            "please call Line.build_tracker() first.")

    assert mode in [None, 'set', 'shift', 'normalized_transverse']
    Particles = xt.Particles  # To get the right Particles class depending on pyheatail interface state

    assert 'at_s' not in kwargs, "at_s is not a valid argument for this function"

    # Deprecation warning for scale_with_transverse_norm_emitt
    if scale_with_transverse_norm_emitt is not None:
        _print('Warning! '
            "scale_with_transverse_norm_emitt is deprecated. Use "
            "nemitt_x and nemitt_y instead.")

    if (particle_ref is not None and particle_on_co is not None):
        raise ValueError("`particle_ref` and `particle_on_co`"
                " cannot be provided at the same time")

    if particle_on_co is None and particle_ref is None:
        if line is not None:
            particle_ref = line.particle_ref

    if particle_ref is None:
        assert particle_on_co is not None, (
            "`particle_ref` or `particle_on_co` must be provided!")
        particle_ref = particle_on_co

    if not isinstance(particle_ref._buffer.context, xo.ContextCpu):
        particle_ref = particle_ref.copy(_context=xo.ContextCpu())

    # Move other input parameters to cpu if needed
    # Generated by:
    # for nn in 'x px y py zeta delta x_norm px_norm y_norm py_norm'.split():
    #     _print(f'{nn} = ({nn}.get() if hasattr({nn}, "get") else {nn})')
    x = (x.get() if hasattr(x, "get") else x)
    px = (px.get() if hasattr(px, "get") else px)
    y = (y.get() if hasattr(y, "get") else y)
    py = (py.get() if hasattr(py, "get") else py)
    zeta = (zeta.get() if hasattr(zeta, "get") else zeta)
    delta = (delta.get() if hasattr(delta, "get") else delta)
    pzeta = (pzeta.get() if hasattr(pzeta, "get") else pzeta)
    ptau = (ptau.get() if hasattr(ptau, "get") else ptau)
    x_norm = (x_norm.get() if hasattr(x_norm, "get") else x_norm)
    px_norm = (px_norm.get() if hasattr(px_norm, "get") else px_norm)
    y_norm = (y_norm.get() if hasattr(y_norm, "get") else y_norm)
    py_norm = (py_norm.get() if hasattr(py_norm, "get") else py_norm)
    zeta_norm = (zeta_norm.get() if hasattr(zeta_norm, "get") else zeta_norm)
    pzeta_norm = (pzeta_norm.get() if hasattr(pzeta_norm, "get") else pzeta_norm)

    if line is not None and line.iscollective and not include_collective:
        logger.warning('Ignoring collective elements in particles generation.')
        line = line._get_non_collective_line()

    # Compute pzeta from delta
    if delta is not None:
        assert pzeta is None
        assert ptau is None
        if not np.isscalar(delta):
            delta = np.array(delta)
        beta0 = particle_ref._xobject.beta0[0]
        delta_beta0 = delta * beta0
        ptau_beta0 = (delta_beta0 * delta_beta0
                            + 2. * delta_beta0 * beta0 + 1.)**0.5 - 1.
        pzeta = ptau_beta0 / beta0 / beta0

    # Compute pzeta from ptau
    if ptau is not None:
        assert pzeta is None
        assert delta is None
        if not np.isscalar(ptau):
            ptau = np.array(ptau)
        pzeta = ptau / particle_ref._xobject.beta0[0]

    if (x_norm is not None or px_norm is not None
            or y_norm is not None or py_norm is not None
            or zeta_norm is not None or pzeta_norm is not None):

        if mode is None:
            mode = 'normalized_transverse'
        else:
            assert mode == 'normalized_transverse'

    if mode is None:
        mode = 'set'

    assert particle_ref._capacity == 1
    ref_dict = {
        'q0': particle_ref.q0,
        'mass0': particle_ref.mass0,
        't_sim': particle_ref.t_sim,
        'p0c': particle_ref.p0c[0],
        'gamma0': particle_ref.gamma0[0],
        'beta0': particle_ref.beta0[0],
        'pdg_id': particle_ref.pdg_id[0],
        'anomalous_magnetic_moment': particle_ref.anomalous_magnetic_moment[0],
    }
    part_dict = ref_dict.copy()

    if at_element is not None or match_at_s is not None:
        # Only this case is covered if not starting at element 0
        assert line is not None
        assert mode == 'normalized_transverse'
        if isinstance(at_element, str):
            at_element = line._element_names_unique.index(at_element)
        assert R_matrix is None # Not clear if it is at the element or at start machine
        if particle_on_co is not None:
            assert particle_on_co._xobject.at_element == 0

    if match_at_s is not None:
        assert at_element is not None, (
            'If `match_at_s` is provided, `at_element` needs to be provided and'
            'needs to correspond to the corresponding element in the sequence'
        )
        s_elements = line.get_s_elements()
        s_at_element = s_elements[at_element]
        if np.abs(match_at_s - s_at_element) < s_tol:
            at_element_line_rmat = at_element
            line_rmat = line
            match_at_s = None
        else:
            # Match at a position where there is no marker and backtrack to the previous marker
            expected_at_element = np.where(np.array(s_elements)<=match_at_s)[0][-1]
            assert at_element == expected_at_element or (
                    at_element < expected_at_element and
                        all([xt._is_aperture(line.element_dict[nn], line)
                            or xt._behaves_like_drift(line.element_dict[nn], line)
                    for nn in line._element_names_unique[at_element:expected_at_element]])), (
                "`match_at_s` can only be placed in the drifts downstream of the "
                "specified `at_element`. No active element can be present in between."
                )
            (tracker_rmat, _
                ) = xt.twiss._build_auxiliary_tracker_with_extra_markers(
                    tracker=line.tracker, at_s=[match_at_s],
                    marker_prefix='xpart_rmat_')
            at_element_line_rmat = tracker_rmat.line._element_names_unique.index(
                                                                    'xpart_rmat_0')
            line_rmat = tracker_rmat.line
    else:
        line_rmat = line
        at_element_line_rmat = at_element

    if mode == 'normalized_transverse':

        if W_matrix is None and line is not None:
            if method is not None:
                kwargs['method'] = method
            tw = line_rmat.twiss(particle_on_co=particle_on_co,
                                    particle_ref=particle_ref,
                                    R_matrix=R_matrix, **kwargs)
            tw_state = tw.get_twiss_init(at_element=
                (at_element_line_rmat if at_element_line_rmat is not None else 0))

            # This is not initialized by get_twiss_init
            tw_state.particle_on_co.at_element = line_rmat._element_names_unique.index(
                                                        tw_state.element_name)

            WW = tw_state.W_matrix
            particle_on_co = tw_state.particle_on_co
        elif W_matrix is None and R_matrix is not None:
            import xtrack.linear_normal_form as lnf
            WW, _, _, _ = lnf.compute_linear_normal_form(R_matrix, **kwargs)
        else:
            WW = W_matrix

        num_particles = _check_lengths(num_particles=num_particles,
            x=x, px=px, y=y, py=py, zeta=zeta, pzeta=pzeta,
            x_norm=x_norm, px_norm=px_norm,
            y_norm=y_norm, py_norm=py_norm,
            zeta_norm=zeta_norm, pzeta_norm=pzeta_norm)

        if scale_with_transverse_norm_emitt is not None:
            assert len(scale_with_transverse_norm_emitt) == 2
            assert nemitt_x is None and nemitt_y is None, (
                "If `scale_with_transverse_norm_emitt` is provided, "
                "`nemitt_x` and `nemitt_y` should not be provided.")

            nemitt_x = scale_with_transverse_norm_emitt[0]
            nemitt_y = scale_with_transverse_norm_emitt[1]

        if nemitt_x is None:
            gemitt_x = 1
        else:
            gemitt_x = (nemitt_x / particle_ref._xobject.beta0[0]
                        / particle_ref._xobject.gamma0[0])

        if nemitt_y is None:
            gemitt_y = 1
        else:
            gemitt_y = (nemitt_y / particle_ref._xobject.beta0[0]
                        / particle_ref._xobject.gamma0[0])

        if nemitt_zeta is None:
            gemitt_zeta = 1
        else:
            gemitt_zeta = (nemitt_zeta / particle_ref._xobject.beta0[0]
                        / particle_ref._xobject.gamma0[0])


        n_constraints = sum([vv is not None for vv in [x, x_norm, px, px_norm,
            y, y_norm, py, py_norm, zeta, zeta_norm, pzeta, pzeta_norm]])

        if n_constraints > 6:
            raise ValueError('Too many constraints. At most 6 among `x`, `x_norm`, '
                                '`px`, `px_norm`, `y`, `y_norm`, `py`, `py_norm`, '
                                '`zeta`, `zeta_norm`, `pzeta`, `pzeta_norm` can be '
                                'provided.')
        while n_constraints < 6:
            if x is None and x_norm is None:
                x_norm = 0
                n_constraints += 1
                continue
            if px is None and px_norm is None:
                px_norm = 0
                n_constraints += 1
                continue
            if y is None and y_norm is None:
                y_norm = 0
                n_constraints += 1
                continue
            if py is None and py_norm is None:
                py_norm = 0
                n_constraints += 1
                continue
            if zeta is None and zeta_norm is None:
                zeta_norm = 0
                n_constraints += 1
                continue
            if pzeta is None and pzeta_norm is None:
                pzeta_norm = 0
                n_constraints += 1
                continue

        # if sum([vv is not None for vv in [x, x_norm, px, px_norm]]) <= 1:
        #     if x is None and x_norm is None:
        #         x_norm = 0
        #     if px is None and px_norm is None:
        #         px_norm = 0

        # if sum([vv is not None for vv in [y, y_norm, py, py_norm]]) <= 1:
        #     if y is None and y_norm is None:
        #         y_norm = 0
        #     if py is None and py_norm is None:
        #         py_norm = 0

        # if sum([vv is not None for vv in [zeta, zeta_norm, pzeta, pzeta_norm]]) <= 1:
        #     if zeta is None and zeta_norm is None:
        #         zeta_norm = 0
        #     if pzeta is None and pzeta_norm is None:
        #         pzeta_norm = 0

        BB = np.zeros(shape=(12, num_particles), dtype=np.float64)
        AA = np.zeros(shape=(12, 12), dtype=np.float64)

        # The first 6 equations are X - WW * X_norm = X_CO
        AA[:6, :6] = np.eye(6)
        AA[:6, 6:] = -WW

        BB[0, :] = particle_on_co._xobject.x[0]
        BB[1, :] = particle_on_co._xobject.px[0]
        BB[2, :] = particle_on_co._xobject.y[0]
        BB[3, :] = particle_on_co._xobject.py[0]
        BB[4, :] = particle_on_co._xobject.zeta[0]
        BB[5, :] = particle_on_co._xobject.ptau[0] / particle_on_co._xobject.beta0[0]

        # The next 6 equations fix either X or X_norm
        i_fill = 6
        for ii, rr in enumerate([x, px, y, py, zeta, pzeta]):
            if rr is not None:
                BB[i_fill, :] = np.array(rr)
                AA[i_fill, ii] = 1
                i_fill += 1

        for ii, (rr_norm,  gemitt) in enumerate(zip(
                    [x_norm, px_norm, y_norm, py_norm, zeta_norm, pzeta_norm],
                    [gemitt_x, gemitt_x, gemitt_y, gemitt_y, gemitt_zeta, gemitt_zeta])):
            if rr_norm is not None:
                rr_norm_scaled = np.sqrt(gemitt) * np.array(rr_norm)
                BB[i_fill, :] = rr_norm_scaled
                AA[i_fill, ii + 6] = 1
                i_fill += 1

        # Solve and extract geometric coordinates
        X_Xhat = np.linalg.solve(AA, BB)
        XX = X_Xhat[:6, :]

    elif mode == 'set':

        if x is None: x = 0
        if px is None: px = 0
        if y is None: y = 0
        if py is None: py = 0
        if zeta is None: zeta = 0
        if pzeta is None: pzeta = 0

        if R_matrix is not None:
            logger.warning('R_matrix provided but not used in this mode!')

        num_particles = _check_lengths(num_particles=num_particles,
            zeta=zeta, delta=delta, x=x, px=px,
            y=y, py=py)

        XX = np.zeros(shape=(6, num_particles), dtype=np.float64)
        XX[0, :] = x
        XX[1, :] = px
        XX[2, :] = y
        XX[3, :] = py
        XX[4, :] = zeta
        XX[5, :] = pzeta

    elif mode == "shift":

        if x is None: x = 0
        if px is None: px = 0
        if y is None: y = 0
        if py is None: py = 0
        if zeta is None: zeta = 0
        if pzeta is None: pzeta = 0

        if R_matrix is not None:
            logger.warning('R_matrix provided but not used in this mode!')

        num_particles = _check_lengths(num_particles=num_particles,
            zeta=zeta, delta=delta, x=x, px=px,
            y=y, py=py)

        XX = np.zeros(shape=(6, num_particles), dtype=np.float64)
        XX[0, :] = x + particle_ref.x
        XX[1, :] = px + particle_ref.px
        XX[2, :] = y + particle_ref.y
        XX[3, :] = py + particle_ref.py
        XX[4, :] = zeta + particle_ref.zeta
        XX[5, :] = pzeta + particle_ref.ptau / particle_ref._xobject.beta0[0]
    else:
        raise ValueError('What?!')

    part_dict['x'] = XX[0, :]
    part_dict['px'] = XX[1, :]
    part_dict['y'] = XX[2, :]
    part_dict['py'] = XX[3, :]
    part_dict['zeta'] = XX[4, :]
    part_dict['ptau'] = XX[5, :] * particle_ref._xobject.beta0[0]

    part_dict['weight'] = np.ones(num_particles, dtype=np.float64)

    if _context is None and _buffer is None and line is not None:
        _context = line._buffer.context

    particles = Particles(_context=_context, _buffer=_buffer, _offset=_offset,
                          _capacity=_capacity,**part_dict)

    particles.particle_id[:num_particles] = particles._buffer.context.nparray_to_context_array(
                                   np.arange(0, num_particles, dtype=np.int64))
    if weight is not None:
        particles.weight[:num_particles] = weight

    if match_at_s is not None:
        # Backtrack to at_element
        length_aux_drift = -match_at_s + line.get_s_position(at_element)
        assert length_aux_drift <= 0
        auxdrift = xt.Drift(length=length_aux_drift,
                            _context=line._buffer.context)
        auxdrift.track(particles)

    if at_element is not None:
        if match_at_s is not None:
            particles.s[:num_particles] = particle_on_co._xobject.s[0] + length_aux_drift
        else:
            assert particle_on_co.at_element[0] == at_element
            particles.s[:num_particles] = particle_on_co._xobject.s[0]
        particles.at_element[:num_particles] = at_element

        particles.start_tracking_at_element = at_element

    particles.spin_x[:num_particles] = particle_ref._xobject.spin_x[0]
    particles.spin_y[:num_particles] = particle_ref._xobject.spin_y[0]
    particles.spin_z[:num_particles] = particle_ref._xobject.spin_z[0]
    if 'spin_x' in kwargs:
        particles.spin_x[:num_particles] = kwargs['spin_x']
    if 'spin_y' in kwargs:
        particles.spin_y[:num_particles] = kwargs['spin_y']
    if 'spin_z' in kwargs:
        particles.spin_z[:num_particles] = kwargs['spin_z']

    return particles



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import numpy as np

import xpart as xp
import xtrack as xt
import xobjects as xo

from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_build_particles_normalized(test_context):
    # Build a reference particle
    p0 = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1, p0c=7e12, x=1, y=3,
                      delta=[10], _context=test_context)


    # Load machine model (from pymask)
    filename = xt._pkg_root.parent.joinpath('test_data/lhc_no_bb/line_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
    line = xt.Line.from_dict(input_data['line'])
    line.build_tracker(_context=test_context)

    # Built a set of three particles with different x coordinates
    particles = xp.build_particles(_context=test_context,
                                   line=line, particle_ref=p0,
                                   zeta=0, delta=1e-3,
                                   x_norm=[1, 0, -1],  # in sigmas
                                   px_norm=[0, 1, 0],  # in sigmas
                                   nemitt_x=3e-6, nemitt_y=3e-6)

    dct = particles.to_dict() # transfers it to cpu
    xo.assert_allclose(dct['x'], [-0.0003883 , -0.0006076 , -0.00082689],
                       rtol=0, atol=1e-7)
    assert np.isclose(dct['ptau'][1], 1e-3, rtol=0, atol=1e-9)
    assert np.isclose(1/(dct['rpp'][1]) - 1, 1e-3, rtol=0, atol=1e-10)
    assert np.all(dct['p0c'] == 7e12)

    # Same as before with R matrix provided as input
    tw = line.twiss(particle_ref=p0)
    R_matrix = tw.R_matrix
    particle_on_co = tw.particle_on_co.copy()
    particles = xp.build_particles(_context=test_context,
                                   particle_on_co=particle_on_co,
                                   R_matrix=R_matrix,
                                   zeta=0, delta=1e-3,
                                   x_norm=[1, 0, -1],  # in sigmas
                                   px_norm=[0, 1, 0],  # in sigmas
                                   nemitt_x=3e-6, nemitt_y=3e-6)
    dct = particles.to_dict() # transfers it to cpu
    xo.assert_allclose(dct['x'], [-0.0003883, -0.0006076, -0.00082689],
                       rtol=0, atol=1e-7)
    assert np.isclose(dct['ptau'][1], 1e-3, rtol=0, atol=1e-9)
    assert np.isclose(1/(dct['rpp'][1]) - 1, 1e-3, rtol=0, atol=1e-10)
    assert np.all(dct['p0c'] == 7e12)

    # Test the 4d mode
    for ee in line.elements:
        if isinstance(ee, xt.Cavity):
            ee.voltage = 0

    # Built a set of three particles with different x coordinates
    particles = xp.build_particles(_context=test_context, method='4d',
                                   line=line, particle_ref=p0,
                                   zeta=0, delta=1e-3,
                                   x_norm=[1,0,-1], # in sigmas
                                   px_norm=[0,1,0], # in sigmas
                                   nemitt_x=3e-6, nemitt_y=3e-6)

    dct = particles.to_dict() # transfers it to cpu
    xo.assert_allclose(dct['x'], [-0.00038813 , -0.00060738 , -0.00082664],
                       rtol=0, atol=1e-7)
    assert np.isclose(dct['ptau'][1], 1e-3, rtol=0, atol=1e-9)
    assert np.isclose(1/(dct['rpp'][1]) - 1, 1e-3, rtol=0, atol=1e-10)
    assert np.all(dct['p0c'] == 7e12)


@for_all_test_contexts
def test_build_particles_normalized_ions(test_context):
    filename = xt._pkg_root.parent / 'test_data' / 'sps_ions' / 'line_and_particle.json'
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
    line = xt.Line.from_dict(input_data)
    p0 = line.particle_ref
    line.build_tracker(_context=test_context)

    # Built a set of three particles with different x coordinates
    particles = xp.build_particles(_context=test_context,
                                   line=line, particle_ref=p0,
                                   zeta=0, delta=1e-5,
                                   x_norm=[1, 0, -1],  # in sigmas
                                   px_norm=[0, 1, 0],  # in sigmas
                                   nemitt_x=3e-6, nemitt_y=3e-6)

    dct = particles.to_dict() # transfers it to cpu
    xo.assert_allclose(dct['x'], [6.5404e-3, 1.21e-5, -6.5163e-3],
                       rtol=0, atol=1e-7)
    assert np.isclose(dct['ptau'][1], 9.906e-6, rtol=0, atol=1e-9)
    assert np.isclose(1/(dct['rpp'][1]) - 1, 1e-5, rtol=0, atol=1e-10)
    xo.assert_allclose(dct['p0c'], 1.4024063e+12, rtol=0, atol=1e3)


@for_all_test_contexts
def test_build_particles_normalized_closed_orbit(test_context):
    # Build a reference particle
    p0 = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1, p0c=7e12, x=1, y=3,
                      delta=[10], _context=test_context)


    # Load machine model (from pymask)
    filename = xt._pkg_root.parent.joinpath('test_data/lhc_no_bb/line_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
    line = xt.Line.from_dict(input_data['line'])
    line.build_tracker(_context=test_context)

    co_guess = xp.build_particles(particle_ref=p0)
    particle_on_co = line.find_closed_orbit(co_guess=co_guess)

    # Built a set of three particles with different x coordinates
    particles = xp.build_particles(_context=test_context,
                                   line=line, particle_ref=p0,
                                   zeta=particle_on_co._xobject.zeta[0],
                                   delta=particle_on_co._xobject.delta[0],
                                   x_norm=0, # in sigmas
                                   px_norm=0, # in sigmas
                                   nemitt_x=3e-6, nemitt_y=3e-6)

    dct = particles.to_dict()
    dct_co = particle_on_co.to_dict()

    for nn in 'x px y py zeta delta ptau rvv rpp gamma0 beta0 p0c'.split():
        xo.assert_allclose(dct[nn], dct_co[nn], atol=1e-15, rtol=0)


@for_all_test_contexts
def test_build_particles_normalized_match_at_s(test_context):
    # Build a reference particle
    p0 = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1, p0c=7e12, x=1, y=3,
                      delta=[10], _context=test_context)

    # Load machine model (from pymask)
    filename = xt._pkg_root.parent.joinpath('test_data/lhc_no_bb/line_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
    line = xt.Line.from_dict(input_data['line'])

    at_element = 'ip6'
    i_start = line.element_names.index(at_element)
    s_start = line.get_s_position(i_start)
    # Find first active element
    for iele in range(i_start, len(line)):
        ee = line[iele]
        if not isinstance(ee, (xt.Drift, xt.Marker)):
            i_next_active = iele
            break

    match_at_s = line.get_s_position(i_next_active) - 0.11

    # Ensure there is a Marker between at_element and match_at_s (to test behave_likes_drift)
    line.insert_element(element=xt.Marker(), name='test_marker', at_s=s_start + 0.3*(match_at_s-s_start))

    line.build_tracker(_context=test_context)

    # Built a set of three particles with different x coordinates
    particles = xp.build_particles(_context=test_context,
                                   line=line, particle_ref=p0,
                                   x=0.02, # in meters
                                   px_norm=np.random.normal(scale=0.1, size=300), # in sigmas
                                   y_norm=np.random.normal(scale=0.1, size=300),  # in sigmas
                                   py_norm=np.random.normal(scale=0.1, size=300), # in sigmas
                                   zeta=np.random.normal(scale=0.05, size=300),   # in meters
                                   delta=np.random.normal(scale=1e-4, size=300),
                                   nemitt_x=3e-6, nemitt_y=3e-6,
                                   at_element=at_element, match_at_s=match_at_s)

    particles.move(_context=xo.context_default)
    assert not np.allclose(particles.x, 0.02, atol=1e-20)
    line.discard_tracker()
    line.insert_element(element=xt.Marker(), name='match_at_s', at_s=match_at_s)
    line.build_tracker(_context=test_context)
    particles.move(_context=test_context)
    line.track(particles, ele_stop='match_at_s')
    particles.move(_context=xo.context_default)
    assert (np.unique(particles.at_element[particles.state>0])[0]
                     == line.element_names.index('match_at_s'))
    xo.assert_allclose(particles.x, 0.02, atol=1e-14, rtol=0)


@for_all_test_contexts
def test_build_perticles_dispersion(test_context):

    # Load machine model (from pymask)
    filename = xt._pkg_root.parent.joinpath(
                                'test_data/lhc_no_bb/line_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)

    line=xt.Line.from_dict(input_data['line'])
    line.particle_ref = xp.Particles(mass0=xp.PROTON_MASS_EV,
                                               q0=1, p0c=7e12)
    line.build_tracker(_context=test_context)

    particles = line.build_particles(nemitt_x=3e-6, nemitt_y=3e-6,
                    x=[1e-3, -1e-3], x_norm=0, px_norm=0, y_norm=0, py_norm=0)


    tw = line.twiss()
    norm_coords = tw.get_normalized_coordinates(
                                    particles, nemitt_x=3e-6, nemitt_y=3e-6)
    particles.move(_context=xo.ContextCpu())

    xo.assert_allclose(particles.x, [1e-3, -1e-3], atol=1e-10, rtol=0)
    xo.assert_allclose((particles.x-tw.x[0])/particles.delta, tw.dx[0],
                       atol=5e-3, rtol=0)

    xo.assert_allclose(norm_coords['x_norm'], 0, atol=1e-12)
    xo.assert_allclose(norm_coords['px_norm'], 0, atol=1e-12)
    xo.assert_allclose(norm_coords['y_norm'], 0, atol=1e-12)
    xo.assert_allclose(norm_coords['py_norm'], 0, atol=1e-12)

    particles = line.build_particles(nemitt_x=3e-6, nemitt_y=3e-6,
                    x=[1e-3, -1e-3], x_norm=[0.3, 0.4], px_norm=[0.5, 0.6],
                    y_norm=[0.7, 0.8], py_norm=[0.9, 1.0])

    tw = line.twiss()
    norm_coords = tw.get_normalized_coordinates(
                                    particles, nemitt_x=3e-6, nemitt_y=3e-6)
    particles.move(_context=xo.ContextCpu())

    xo.assert_allclose(particles.x, [1e-3, -1e-3], atol=1e-10, rtol=0)

    xo.assert_allclose(norm_coords['x_norm'], [0.3, 0.4], 1e-12)
    xo.assert_allclose(norm_coords['px_norm'], [0.5, 0.6], 1e-12)
    xo.assert_allclose(norm_coords['y_norm'], [0.7, 0.8], 1e-12)
    xo.assert_allclose(norm_coords['py_norm'], [0.9, 1.0], 1e-12)



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xpart as xp

from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_merge(test_context):
    p1 = xp.Particles(x=[1, 2, 3], delta=[1e-3, 2e-2, 3e-3], p0c=7e12,
                      mass0=xp.ELECTRON_MASS_EV,
                      _context=test_context)
    p2 = xp.Particles(x=[4, 5], p0c=7e12, mass0=xp.ELECTRON_MASS_EV)
    p3 = xp.Particles(x=6, delta=[-1e-1], p0c=7e12, mass0=xp.ELECTRON_MASS_EV)

    particles = xp.Particles.merge([p1,p2,p3])

    assert particles._buffer.context == test_context
    assert particles.mass0 == xp.ELECTRON_MASS_EV
    tocpu = test_context.nparray_from_context_array
    for nn in 'x px y py zeta delta ptau rpp rvv gamma0 p0c'.split():
        assert np.all(tocpu(getattr(particles, nn)[0:3]) ==
                      p1.to_dict()[nn])
        assert np.all(tocpu(getattr(particles, nn)[3:5]) ==
                      p2.to_dict()[nn])
        assert np.all(tocpu(getattr(particles, nn)[5:]) ==
                      p3.to_dict()[nn])



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xobjects as xo
import xpart as xp

from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts(excluding='ContextPyopencl')
def test_add_particles(test_context):
    capacity = 10

    p_base = xp.Particles(x=[1, 2, 3], delta=[1e-3, 2e-2, 3e-3], p0c=7e12,
                          _capacity=capacity,
                          mass0=xp.ELECTRON_MASS_EV,
                         _context=test_context)
    
    p1 = p_base.copy()
    p2 = p_base.copy()
    
    p1._init_random_number_generator()

    custom_seeds = np.linspace(10, 20, capacity).astype(int)
    p2._init_random_number_generator(seeds=custom_seeds)

    p_add = xp.Particles(
        x=[4, 5], p0c=7e12, mass0=xp.ELECTRON_MASS_EV, _context=test_context)
    # To test this doesn't override the base seeds
    p_add._init_random_number_generator(seeds=[60, 70])

    p1_check = p1.copy()
    p2_check = p2.copy()
    p1.add_particles(p_add, keep_lost=False)
    p2.add_particles(p_add, keep_lost=False)


    tocpu = test_context.nparray_from_context_array
    # Check parameters that need to be updated
    for nn in 'x px y py zeta delta ptau rpp rvv gamma0 p0c'.split():
        assert np.all(tocpu(getattr(p1, nn)[0:3]) ==
                      p_base.to_dict()[nn][0:3])
        assert np.all(tocpu(getattr(p1, nn)[3:5]) ==
                      p_add.to_dict()[nn][0:2])
        
    # Check that the seeds (allocated up to capacity) are not overwritten
    for nn in '_rng_s1 _rng_s2 _rng_s3'.split():
        assert np.all(tocpu(getattr(p1, nn)) ==
                      p1_check.to_dict()[nn])
        
        assert np.all(tocpu(getattr(p2_check, nn)) ==
                      p2_check.to_dict()[nn])


# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xpart as xp
import xobjects as xo

from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_build_particles_shift(test_context):
    for ctx_ref in [test_context, None]:
        # Build a reference particle
        p0 = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1, p0c=7e12, x=1, y=3,
                          delta=[1e-4], _context=ctx_ref)


        # Built a set of three particles with different x coordinates
        particles = xp.build_particles(mode='shift', particle_ref=p0, y=[1,2,3],
                                       _context=test_context)

        dct = particles.to_dict() # transfers it to cpu
        assert np.isclose(dct['ptau'][1], 1e-4, rtol=0, atol=1e-9)
        assert np.isclose(1/(dct['rpp'][1]) - 1, 1e-4, rtol=0, atol=1e-14)
        assert np.all(dct['p0c'] == 7e12)
        assert dct['x'][1] == 1.0
        assert dct['y'][1] == 5.0



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xpart as xp
import xobjects as xo

from xobjects.test_helpers import for_all_test_contexts


@for_all_test_contexts
def test_copy(test_context):
    p1 = xp.Particles(x=[1, 2, 3], delta=1e-3, _context=test_context)

    # Make a copy of p1 in the same context
    p2 = p1.copy()

    # Copy across contexts
    p3 = p1.copy(_context=xo.ContextCpu())

    # And back
    p4 = p3.copy(_context=test_context)

    dct1 = p1.to_dict()
    dct2 = p2.to_dict()
    dct3 = p3.to_dict()
    dct4 = p4.to_dict()
    for nn in 'x px y py zeta delta ptau rpp rvv gamma0 p0c'.split():
        for dct in [dct2, dct3, dct4]:
            assert np.all(dct[nn] == dct1[nn])



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import numpy as np

import xpart as xp

def test_generate_2D_uniform_circular_sector():

    (x_norm, px_norm, r_points, theta_points
            ) = xp.generate_2D_uniform_circular_sector(
            num_particles=10000, r_range=(0.6, 0.9),
            theta_range=(0.25*np.pi, 1.75*np.pi))

    assert (len(x_norm) == len(px_norm) == len(r_points)
            == len(theta_points) == 10000)

    assert np.isclose(np.max(np.abs(x_norm)), 0.9, rtol=1e-2)
    assert np.isclose(np.max(px_norm), 0.9, rtol=1e-2)
    assert np.isclose(np.min(np.sqrt(x_norm**2 + px_norm**2)), 0.6,
                      rtol=1e-2)

    assert np.isclose(np.min(theta_points), 0.25*np.pi, rtol=1e-2)
    assert np.isclose(np.max(theta_points), 1.75*np.pi, rtol=1e-2)
    assert np.isclose(np.min(r_points), 0.6, rtol=1e-2)
    assert np.isclose(np.max(r_points), 0.9, rtol=1e-2)




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json
from itertools import product
import numpy as np

import xpart as xp
import xtrack as xt
import xobjects as xo
from xobjects.test_helpers import for_all_test_contexts, fix_random_seed


@for_all_test_contexts
@fix_random_seed(7363444)
def test_pencil_with_absolute_cut(test_context):

    num_particles = 10000
    nemitt_x = 2.5e-6
    nemitt_y = 3e-6

    # Load machine model
    filename = xt._pkg_root.parent.joinpath('test_data/hllhc15_noerrors_nobb/line_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
    line=xt.Line.from_dict(input_data['line'])
    line.build_tracker(_context=test_context)
    line.particle_ref = xp.Particles.from_dict(input_data['particle'])

    # Location of the collimator
    at_element = 'tcp.6l3.b1' # High dispersion
    # at_element = 'tcp.d6l7.b1' # Low dispersion
    at_s = line.get_s_position(at_element) + 1.
    absolute_cut = 3e-3 # position of the jaw
    pencil_dr_sigmas = 3 # width of the pencil

    for plane, side in product(['x', 'y'], ['+', '-']):

        if side == '+':
            absolute_cut = 3e-3
        else:
            absolute_cut = -3e-3

        print(f'plane={plane} side={side}')

        # generate pencil beam in absolute coordinates
        v_absolute, pv_absolute = xp.generate_2D_pencil_with_absolute_cut(num_particles,
                            plane=plane, absolute_cut=absolute_cut, dr_sigmas=pencil_dr_sigmas,
                            side=side, line=line,
                            nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                            at_element=at_element, match_at_s=at_s)

        # Vertical plane: generate gaussian distribution in normalized coordinates
        w_in_sigmas, pw_in_sigmas = xp.generate_2D_gaussian(num_particles)

        # Longitudinal plane: generate gaussian distribution matched to bucket
        zeta, delta = xp.generate_longitudinal_coordinates(
                num_particles=num_particles, distribution='gaussian',
                sigma_z=10e-2, line=line)

        # Combine the three planes
        # (the normalized coordinates in y/py are blurred in order to preserve the geometric ones)
        if plane == 'x':
            particles = line.build_particles(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                            x=v_absolute, px=pv_absolute,
                            y_norm=w_in_sigmas, py_norm=pw_in_sigmas,
                            zeta=zeta, delta=delta,
                            at_element=at_element, match_at_s=at_s)
        elif plane == 'y':
            particles = line.build_particles(nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                            x_norm=w_in_sigmas, px_norm=pw_in_sigmas,
                            y=v_absolute, py=pv_absolute,
                            zeta=zeta, delta=delta,
                            at_element=at_element, match_at_s=at_s)

        # Drift to at_s position for checking
        drift_to_at_s = xt.Drift(_context=test_context,
                        length=at_s-line.get_s_position(at_element))
        drift_to_at_s.track(particles)

        particles.move(_context=xo.ContextCpu())
        # Checks

        tw_at_s = line.twiss(at_s=at_s)
        norm_coords = tw_at_s.get_normalized_coordinates(
                        particles, nemitt_x=nemitt_x, nemitt_y=nemitt_y,
                        _force_at_element=0)

        v = getattr(particles, plane)
        pv = getattr(particles, 'p'+plane)
        betv = getattr(tw_at_s, 'bet'+plane)[0]
        alfv = getattr(tw_at_s, 'alf'+plane)[0]
        nemitt_v = {'x': nemitt_x, 'y': nemitt_y}[plane]
        sigma_v = np.sqrt(betv*nemitt_v
            /particles._xobject.beta0[0]/particles._xobject.gamma0[0])

        assert(np.isclose(np.min(np.abs(v)), abs(absolute_cut), atol=1e-7))
        assert(np.isclose(np.max(np.abs(v)), abs(absolute_cut) + sigma_v*pencil_dr_sigmas,
            rtol=1e-3, atol=0))

        i_tip = np.argmax(np.abs(v))
        assert np.isclose(pv[i_tip]/v[i_tip], -alfv/betv, atol=5e-4)

        if side == '+':
            assert np.all(v >= 0)
        else:
            assert np.all(v <= 0)

        other_plane = {'x': 'y', 'y': 'x'}[plane]
        w_norm = getattr(norm_coords, other_plane+'_norm')
        pw_norm = getattr(norm_coords, 'p'+other_plane+'_norm')

        xo.assert_allclose(w_in_sigmas, w_norm, atol=1e-12, rtol=1e-7)
        xo.assert_allclose(pw_in_sigmas, pw_norm, atol=1e-12, rtol=1e-7)

        xo.assert_allclose(zeta, particles.zeta, atol=1e-12, rtol=1e-7)
        xo.assert_allclose(delta, particles.delta, atol=1e-12, rtol=1e-7)



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from xtrack.linear_normal_form import healy_symplectify, compute_linear_normal_form




# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from setuptools import setup, find_packages, Extension
from pathlib import Path

#######################################
# Prepare list of compiled extensions #
#######################################

extensions = []


#########
# Setup #
#########

version_file = Path(__file__).parent / 'xpart/_version.py'
dd = {}
with open(version_file.absolute(), 'r') as fp:
    exec(fp.read(), dd)
__version__ = dd['__version__']

setup(
    name='xpart',
    version=__version__,
    description='Generation of Particle Ensembles',
    long_description=('Tools for the generation of particle ensembles'
                'in beam dynamics simulations.\n\n'
                'This package is part of the Xsuite collection.'),
    url='https://xsuite.readthedocs.io/',
    packages=find_packages(),
    ext_modules=extensions,
    include_package_data=True,
    install_requires=[
        'numpy>=1.0',
        'scipy',
        'xobjects',
        'xtrack',
        ],
    author='G. Iadarola et al.',
    license='Apache 2.0',
    download_url="https://pypi.python.org/pypi/xpart",
    project_urls={
            "Bug Tracker": "https://github.com/xsuite/xsuite/issues",
            "Documentation": 'https://xsuite.readthedocs.io/',
            "Source Code": "https://github.com/xsuite/xpart",
        },
    extras_require={
        'tests': ['cpymad', 'PyHEADTAIL', 'pytest'],
        },
    )



from xpart.longitudinal import generate_longitudinal_coordinates
import numpy as np
from ..general import _print

def generate_binomial_longitudinal_coordinates(num_particles=None,
                                               nemitt_x=None,
                                               nemitt_y=None,
                                               sigma_z=None,
                                               particle_ref=None,
                                               tracker=None,
                                               line=None,
                                               return_matcher=False,
                                               m=5.0
                                               ):

    """
    Function to generate a binomial longitudinal distribution

    Parameters:
    -----------
    num_particles : int
        number of macroparticles
    nemitt_x : float
        normalized horizontal emittance in m rad
    nemitt_y : float
        normalized vertical emittance in m rad
    sigma_z : float
        bunch length in meters
    particle_ref : xp.particle
        reference particle
    tracker : xt.tracker
    line: xt.line
    return_matcher : bool
        whether to also return xp.SingleRFHarmonicMatcher object
    m : float
        binomial parameter, determines fatness of tails. 5.0 is typical value of Pb ions at extraction

    Returns:
    --------
    zeta : np.ndarray
        longitudinal coordinates zeta for particles
    delta : np.ndarray
        relative momentum offset coordinates for particles
    matcher : xp.SingleRFHarmonicMatcher
        RF matcher object
    """

    if line is not None and tracker is not None:
        raise ValueError(
            'line and tracker cannot be provided at the same time.')

    if tracker is not None:
        _print('Warning! '
            "The argument tracker is deprecated. Please use line instead.")
        line = tracker.line

    if line is not None:
        assert line.tracker is not None, ("The line must have a tracker, "
            "please call Line.build_tracker() first.")

    if num_particles is None:
        raise ValueError(
                'Number of particles must be provided')
    if sigma_z is None:
        raise ValueError(
                'Bunch length sigma_z must be provided')

    if particle_ref is None:
        raise ValueError(
                'Reference particle must be provided')

    # If emittances are not provided, set them to default value of one
    if nemitt_x is None:
        nemitt_x = 1.0

    if nemitt_y is None:
        nemitt_y = 1.0

    # Generate longitudinal coordinates s
    zeta, delta, matcher = generate_longitudinal_coordinates(line=line, distribution='binomial',
                            num_particles=num_particles,
                            engine='single-rf-harmonic', sigma_z=sigma_z,
                            particle_ref=particle_ref, return_matcher=True, m=m)

    if return_matcher:
        return zeta, delta, matcher
    else:
        return zeta, delta



# copyright ############################### #
# This file is part of the Xpart Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

import json

import numpy as np

import xpart as xp
import xtrack as xt
import xobjects as xo

from xpart.longitudinal import generate_qgaussian_longitudinal_coordinates

from xobjects.test_helpers import fix_random_seed, for_all_test_contexts


@for_all_test_contexts
@fix_random_seed(4237638)
def test_build_particles_qgaussian(test_context):
    # Build a reference particle
    p0 = xp.Particles(mass0=xp.PROTON_MASS_EV, q0=1, p0c=7e12, x=1, y=3,
                      delta=[10], _context=test_context)

    # Parameters for the test
    num_part = 3000000

    # Load machine model (from pymask)
    filename = xt._pkg_root.parent.joinpath('test_data/lhc_no_bb/line_and_particle.json')
    with open(filename, 'r') as fid:
        input_data = json.load(fid)
    line = xt.Line.from_dict(input_data['line'])
    line.build_tracker(_context=test_context)
	
    # Built a set of three particles with different x coordinates
    zeta, delta, matcher = generate_qgaussian_longitudinal_coordinates(
        num_particles=num_part,
        nemitt_x=3e-6,
        nemitt_y=3e-6,
        sigma_z=0.05,
        particle_ref=p0,
        line=line,
        return_matcher=True,
    )

    # Test if longitudinal coordinates match with Single
	# Generate distribution from RF matcher
    p0.move(_context=xo.ContextCpu())
    tau = zeta / p0.beta0[0]
    tau_distr_y = matcher.tau_distr_y
    tau_distr_x = matcher.tau_distr_x
    dx = tau_distr_x[1] - tau_distr_x[0]
    hist, edges = np.histogram(tau,
                range=(tau_distr_x[0]-dx/2., tau_distr_x[-1]+dx/2.),
                bins=len(tau_distr_x))
    hist = hist / sum(hist) * sum(tau_distr_y)

    assert np.all(np.isclose(hist, tau_distr_y, atol=5.e-2, rtol=1.e-2))
